#!/bin/sh
#
# Prepare a tree for either guest or host use.

# usage: prepare linux-sourcetree [guest|host]

Usage() {
    echo >&2 "Usage: `basename $0` linux-sourcetree [guest|host] [vmmpath]"
    exit 1
}

if [ $# -ne 2 -a $# -ne 3 ]
then
    Usage
fi


quilt="`which quilt`"
symlinks="`which symlinks`"

# Ask a yes-no question and return TRUE for a Yes answer, false for 
# a NO answer.
ask() {
    echo -n "$1 (y/n) ? "
    while read ans
    do
      expr "$ans" : '[Yy].*' && return 0
      expr "$ans" : '[Nn].*' && return 1
      echo -n "$1 (y/n) ? "
    done
}


# Shorten symlink and convert to relative
# Relies on the perl symlinks script.
# Does nothing if that's not installed.
fix_symlink() {
    [ "symlinks" ] && {
	# Have to run symlinks twice to fix fully,
	# once to convert absolute->relative, and once
	# to shorten the resulting link.
	$symlinks -cs $1 > /dev/null 2>&1
	$symlinks -cs $1 > /dev/null 2>&1
    }
}

# Find kernel version from top four lines of kernel Makefile.
kversion() {
    # Forget about EXTRAVERSION?
    (
	eval `sed -n -e 's/ //gp' -e 4q Makefile`
	echo $VERSION.$PATCHLEVEL.$SUBLEVEL$EXTRAVERSION
    )
}

#
# Symlink patches directory and update symlinks into guest.
# Try for versioned patchset first, then generic 
# (recent head-of-tree) patchset.
fix_guest() {

    [ -f "$VMMPATH/userspace/vmm-offsets.h" ] || {
	echo "You need to build vmm-offsets first"
	ask "Do now?" || {
	    echo Cannot continue.
	    exit 1;
	}
	( 
	    cd "$VMMPATH/userspace" 
	    [ -f Makefile ] || {
		    echo >&2 "Please do KERNEL=path-to-host-kernel ./autogen.sh"
		    echo >&2 "Then retry"
		    exit 1
		}
	    make
	) || exit 1
    }

    # Link in the patches directory
    cd "$1"
    V=`kversion`
    if [ -d "$VMMPATH/patches/patches-guest-$V" ] 
    then
	ln -s "$VMMPATH/patches/patches-guest-$V" patches
    else
	ln -s "$VMMPATH/patches/patches-guest" patches
	echo >&2 "Exact patch match for $V not found; trying generic patchset"
    fi
    fix_symlink patches

    # Create the symlinks into the VMM sourcetree for burn.S and vmm-offsets.h

    ( cd include/asm-ia64; ln -fs "${VMMPATH}/userspace/vmm-offsets.h" . ; fix_symlink vmm-offsets.h ; )
    ( cd arch/ia64/kernel; ln -fs "${VMMPATH}/afterburner/burn.S" .; fix_symlink burn.S ; )

    # Apply the patchset
    "$quilt" push -a

    # and remind the user about how to build the guest OS
    cat <<EOF
    Build guest kernel with:
       make EXTRA=-B$VMMPATH/afterburner EXTRA_AFLAGS=-B$VMMPATH/afterburner boot
    Sample configuration at $VMMPATH/configs/config-guest
EOF
}

#
# Symlink patches directory and update symlinks into host.
# Try for versioned patchset first, then generic 
# (recent head-of-tree) patchset.
#
fix_host() {
    cd "$1"
    V=`kversion`
    if [ -d "$VMMPATH/patches/patches-host-$V" ] 
    then
	ln -s "$VMMPATH/patches/patches-host-$V" patches
    else
	ln -s "$VMMPATH/patches/patches-host" patches
	echo >&2 "Exact patch match for $V not found; trying generic patchset"
    fi
    fix_symlink patches
    $quilt push -a
}

vmmd="${3:-`dirname $0`/..}"
VMMPATH="`(cd $vmmd; pwd; )`"

#
# Check VMM seems OK
#
[ -d "$VMMPATH"/patches -a -d "$VMMPATH"/userspace -a -d "$VMMPATH"/afterburner ] || {
    echo >&2 "Can't find a vmm tree in $VMMPATH ... please specify path to base of vmm tree as third argument"
    Usage
    exit 1
}

[ -f "$1/Makefile" -a -d "$1/kernel" -a -d "$1/include/linux" ] || {
    echo >&2 "$1 doesn't look like a kernel source tree."
    Usage
}

[ -e "$1/patches" ] && {
    echo >&2 "$1 already contains a quilt patches directory"
    echo >&2 "Cannot continue: please move it out of the way"
    exit 1
}


[ "$quilt" ] || {
    echo >&2 This script needs quilt.  Please install it.
    exit 1
}

case "$2" 
in
    host)
	fix_host "$1" && {
	    ask "Configure vmm to build against this kernel?" && {
		K=`pwd`;
		[ -d build/ia64 ] && {
			dir="`dirname $K`"
			kern="`basename $K`"
			K="$dir/build/ia64/$kern"
		}
		cd "$VMMPATH"/userspace
		touch AUTHORS ChangeLog NEWS README
		aclocal
		automake -a
		autoconf
		KERNEL="$K" ./configure --host=ia64-linux-gnu
	    }
	}
	;;
    guest)
	fix_guest "$1"
	;;
    *) Usage
       ;;
esac
