/*
 * Asynchronous I/O.
 * For Linux, we can't be told when A/I/O completes.
 * So create a separate I/O process, and signal when complete.
 */
#define _LARGEFILE64_SOURCE
#define _GNU_SOURCE

#include <unistd.h>
#include <stdlib.h>
#include <semaphore.h>
#include <atomic_ops.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include "vm.h"
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <sys/mman.h>
#include "debug.h"


#define MAX_SGE 256


enum status {
	EMPTY = 0,
	READY,
	IN_PROGRESS,
	COMPLETE
};

struct disk_request {
	unsigned seqno;
	enum status status;
	short fd;
	short nreq;
	pid_t pid;
	int   is_write;
	off_t offset;
	struct disk_req req[MAX_SGE];
	unsigned count; /* result */
};


#define __shared_data __attribute__ ((section (".shared")))

#define NREQ 64

AO_t  __shared_data initted;
sem_t __shared_data empty;	/* counts empty slots */
sem_t __shared_data full;	/* counts full slots */
sem_t __shared_data complete;	/* to signal completion for old-style wait_completions */
AO_t __shared_data wr;
unsigned __shared_data rd;
int __shared_data simscsi_irq;
int __shared_data last_quiescent_req;
int __shared_data network_fd;
pid_t __shared_data mainpid;

int  __shared_data do_exit;
struct disk_request __shared_data reqs[NREQ];

extern char VM_filename[];


static void do_request(struct disk_request *rq)
{
	struct disk_req *req = rq->req;
	struct iovec iov[16];
	int i;
#if O_DIRECT != 0
	int redo_direct = 0;
#endif

	IO_DEBUG("Reading from FD %d at offset %llx, req %u, seq %u\n",
		 rq->fd, (unsigned long long)rq->offset, rd, rq->seqno);

	if (rq->offset < 0) {
		fprintf(stderr,"SSC_RDWR at offset %lx\n", rq->offset);
		rq->count = -1;
		return;
	}

	rq->count = 0;
#if O_DIRECT != 0
	if ((rq->offset & 0x1ff) || (req[0].addr & 0x1ff) || req[0].len & 0x3ff) {
		fcntl(rq->fd, F_SETFL, O_NOATIME);
		redo_direct=1;

	}
#endif
	if (lseek(rq->fd, (off_t)rq->offset, SEEK_SET) != rq->offset) {
		vm_perror("lseek");
		rq->count = -1;
		return;
	}


	while (rq->nreq) {
		int max = rq->nreq > 16 ? 16 : rq->nreq;
		int result;

		for (i = 0; i < max; i++) {
#ifdef DEBUG_IO
			fprintf(stderr, "SSC_RDWR: at %p, size %X\n",
				(void *)req[i].addr, req[i].len);
#endif
			iov[i].iov_base = (void *)req[i].addr;
			iov[i].iov_len = req[i].len;
		}

		while ((result = rq->is_write ?
			writev(rq->fd, iov, max)
			:
			readv(rq->fd, iov, max)) == -1)
		{
#ifdef DSM
			if (errno != EFAULT)
			{
#endif
				vm_perror("rdwr");
				break;
#ifdef DSM
			}
			mm_fetch(sc, req->addr & ~PAGE_MASK, 1, req->addr);
#endif
		}
		if (result == -1)
			break;
		rq->nreq -= max;
		req += max;
		rq->count += result;
	}
#if O_DIRECT != 0
	if (redo_direct) {
		fcntl(rq->fd, F_SETFL, O_DIRECT|O_NOATIME);
		redo_direct = 0;
	}
#endif
}


static void doio(void)
{
	for (;;) {
		IO_DEBUG("doio: Waiting on &full\n");
		int oldrd;
		int n;

		sem_wait(&full);
		IO_DEBUG("doio: Done wait &full; rd = %u\n", rd);

		if (do_exit)
			exit(0);
		oldrd = rd;
		if (reqs[rd].status != READY)
			vm_panic(NULL, "AIO queue lost synchronisation, reqs[%d].status = %d\n", rd, reqs[rd].status);
		reqs[rd].status = IN_PROGRESS;
		do_request(&reqs[rd]);
		reqs[rd].status = COMPLETE;

		if (++rd == NREQ)
			rd = 0;
		if (simscsi_irq && reqs[oldrd].seqno) {
			sem_getvalue(&full, &n);
			if (n == 0) {
				union sigval s;
				s.sival_ptr = NULL;
				last_quiescent_req = reqs[oldrd].seqno;
				IO_DEBUG("Interrupt for seqno %d to pid %u\n",
					 reqs[oldrd].seqno, reqs[oldrd].pid);
				if (sigqueue(reqs[oldrd].pid, SIGSCSI, s) == -1) {
					fprintf(stderr, "kill %u with SIGSCSI", reqs[oldrd].pid);
					vm_perror(" failed");
				}
			}
			else IO_DEBUG("Queue non-empty (%d), going around again\n", n);
		}
		else
			sem_post(&complete);

		sem_post(&empty);

	}
}

int queueio(int fd, int seqno, int nreq, off_t offset, int is_write, struct disk_req *req)
{
	struct disk_request *rq;
	unsigned thisreq;
	int value;

	sem_getvalue(&empty, &value);
	IO_DEBUG("queueio: seqno %d, fd %d empty %d\n", seqno, fd, value);
	sem_wait(&empty);

	thisreq = AO_fetch_and_add1(&wr) % NREQ;

	IO_DEBUG("Adding seq %d at %d.  Addr %lx, len %u\n", seqno, thisreq,
		 req->addr, req->len);
	rq = &reqs[thisreq];
	memcpy(rq->req, req, (sizeof *req) * nreq);
	rq->fd = fd;
	rq->offset = offset;
	rq->is_write = is_write;
	rq->seqno = seqno;
	rq->nreq = nreq;
	rq->pid = getpid();
	rq->status = READY;
	IO_DEBUG("queueio: signalling FULL\n");
	sem_post(&full);
	return 0;
}


/*
 * AIOinit is to be called on all CPUs.
 */
void
aio_init(int keep_files)
{
	extern char start_shared_page[];
	extern char end_shared_page[];
	char filename[36];
	unsigned long pagesize = getpagesize();
	int fd;
	void *p;
	int x;
	size_t sz = end_shared_page - start_shared_page;

	sz = (sz + pagesize - 1) & ~ (pagesize - 1);

	IO_DEBUG("aio_init: sizes is %lu\n", (unsigned long)sz);
	sprintf(VM_filename, "/tmp/vm_mem.%d", mainpid);
	sprintf(filename, "%s-io", VM_filename);

	fd = open(filename, O_CREAT|O_RDWR, 0600);
	if (fd == -1) {
		vm_perror(filename);
		exit(1);
	}
	ftruncate(fd, sz);

	p = mmap(start_shared_page, sz, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_SHARED, fd, 0);
	if (p == MAP_FAILED) {
		vm_perror("mapping shared IO page");
		exit(1);
	}
	close(fd);

	/*
	 * Has a different virtual CPU already initialised?
	 */
	if ((x = AO_fetch_and_add1(&initted)) == 0) {
		/* Nope, we're the first CPU to get here */
		rd = 0;
		AO_store(&wr, 0);
		if (sem_init(&empty, 1, NREQ) == -1) {
			vm_perror("sem_init(empty)\n");
			sem_init(&empty, 0, NREQ);
		}
		sem_init(&full, 0, 0);
		sem_init(&complete, 0, 0);
		if (vm_clone() == 0) {
			doio();
		}
	}
	if ((x == (ncpu - 1)) && !keep_files)
		unlink(filename);
}

void aio_shutdown(void)
{
	do_exit = 1;
	sem_post(&full);
}


int aio_getcompletion(void)
{
	IO_DEBUG("aio_getcompletion: Returning %d\n", last_quiescent_req);
	return last_quiescent_req;
}


int aio_wait_completion(struct disk_stat *ds)
{
	int x;

	IO_DEBUG("aio_wait_completion: waiting on semaphor\n");
       	sem_wait(&complete);

	x = rd - 1 ;
	if (x < 0)
		x = NREQ - 1;

	IO_DEBUG("aio_wait_completion.  Req %d, status %d, count %d\n", x, reqs[x].status, reqs[x].count);

	if (reqs[x].status != COMPLETE) {
		ds->count = 0;
		return -1;
	}
	ds->fd = reqs[x].fd;
	ds->count = reqs[x].count;
	reqs[x].status = EMPTY;
	return 0;
}
