/* 
   This file has one non-static funtion which returns the address of a
   linux boot parameters structure full of environment data.
   
   Much of this stuff is taken from the Linux source code and is copyright
   to its respective owners.
*/

#include "ssc.h"
#include "vm.h"
#include "efi.h"
#include "sal.h"
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

typedef struct __ia64_boot_param_t ia64_boot_param_t;

struct __ia64_boot_param_t {
	u64 command_line;	/* physical address of command line
				   arguments */
	u64 efi_systab;		/* physical address of EFI system table */
	u64 efi_memmap;		/* physical address of EFI memory map */
	u64 efi_memmap_size;	/* size of EFI memory map */
	u64 efi_memdesc_size;	/* size of an EFI memory map descriptor
				 */
	u32 efi_memdesc_version; /* memory descriptor version */
	struct {
		u16 num_cols;	/* number of columns on console output device */
		u16 num_rows;	/* number of rows on console output device */
		u16 orig_x;	/* cursor's x position */
		u16 orig_y;	/* cursor's y position */
	} console_info;
	u64 fpswa;		/* physical address of the fpswa interface */
	u64 initrd_start;
	u64 initrd_size;
};

/* 
   Allocate from the BSS. 
   It seems like a better idea than the heap for some reason. 
*/
static char blob[0x10000] __attribute__ ((section(".shared")));
static void *bloballoc(unsigned int size)
{
	static void *p = blob;
	void *np = p;
	p += size;
	return np;
}

extern void pal_emulator_static(void);

static efi_status_t efi_unimplemented(void)
{
	return EFI_UNSUPPORTED;
}

#define SECS_PER_HOUR   (60 * 60)
#define SECS_PER_DAY    (SECS_PER_HOUR * 24)

/* 
 * Compute the `struct tm' representation of *T,
 *  offset OFFSET seconds east of UTC,
 *  and store year, yday, mon, mday, wday, hour, min, sec into *TP.
 *  Return nonzero if successful.  
 */

static int offtime(unsigned long t, efi_time_t * tp)
{


	const unsigned short int __mon_yday[2][13] = {
		/* Normal years.  */
		{0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334,
		 365},
		/* Leap years.  */
		{0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366}
	};
	long int days, rem, y;
	const unsigned short int *ip;

	days = t / SECS_PER_DAY;
	rem = t % SECS_PER_DAY;
	while (rem < 0) {
		rem += SECS_PER_DAY;
		--days;
	}
	while (rem >= SECS_PER_DAY) {
		rem -= SECS_PER_DAY;
		++days;
	}
	tp->hour = rem / SECS_PER_HOUR;
	rem %= SECS_PER_HOUR;
	tp->minute = rem / 60;
	tp->second = rem % 60;
	/* January 1, 1970 was a Thursday.  */
	y = 1970;

#define DIV(a, b) ((a) / (b) - ((a) % (b) < 0))
#define LEAPS_THRU_END_OF(y) (DIV (y, 4) - DIV (y, 100) + DIV (y, 400))
#define __isleap(year)							\
	((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

	while (days < 0 || days >= (__isleap(y) ? 366 : 365)) {
		/* Guess a corrected year, assuming 365 days per year.  */
		long int yg = y + days / 365 - (days % 365 < 0);

		/* Adjust DAYS and Y to match the guessed year.  */
		days -= ((yg - y) * 365 + LEAPS_THRU_END_OF(yg - 1)
			 - LEAPS_THRU_END_OF(y - 1));
		y = yg;
	}
	tp->year = y;
	ip = __mon_yday[__isleap(y)];
	for (y = 11; days < (long int) ip[y]; --y)
		continue;
	days -= ip[y];
	tp->month = y + 1;
	tp->day = days + 1;
	return 1;
}

static efi_status_t fw_efi_get_time(efi_time_t * tm, efi_time_cap_t * tc)
{
	struct timeval tv32bits;

	gettimeofday(&tv32bits, NULL);

	memset(tm, 0, sizeof *tm);
	offtime(tv32bits.tv_sec, tm);
	tm->nanosecond = tv32bits.tv_usec * 1000;
	if (tc)
		memset(tc, 0, sizeof *tc);

	return EFI_SUCCESS;
}

static void
efi_reset_system(int reset_type, efi_status_t status,
		 unsigned long data_size, efi_char16_t * data)
{
	ssc(0, 0, 0, 0, SSC_EXIT);
}

static struct sal_ret_values
sal_emulator(long index, unsigned long in1, unsigned long in2,
	     unsigned long in3, unsigned long in4, unsigned long in5,
	     unsigned long in6, unsigned long in7)
{
	long r9 = 0;
	long r10 = 0;
	long r11 = 0;
	long status;

	/*
	 * Don't do a "switch" here since that gives us code that
	 * isn't self-relocatable.
	 */
	status = 0;
	if (index == SAL_FREQ_BASE) {
		switch (in1) {
		case SAL_FREQ_BASE_PLATFORM:
			r9 = cpu_freq;
			break;

		case SAL_FREQ_BASE_INTERVAL_TIMER:
			/*
			 * Is this supposed to be the cr.itc frequency
			 * or something platform specific?  The SAL
			 * doc ain't exactly clear on this...
			 */
			r9 = 700000000;
			break;

		case SAL_FREQ_BASE_REALTIME_CLOCK:
			r9 = 1;
			break;

		default:
			status = -1;
			break;
		}
	} else if (index == SAL_SET_VECTORS) {
		if (in1 == SAL_VECTOR_OS_BOOT_RENDEZ)
			ssc(in2, in3, 0, 0, SSC_SET_RENDEZVOUS);
	} else if (index == SAL_GET_STATE_INFO) {
		;
	} else if (index == SAL_GET_STATE_INFO_SIZE) {
		;
	} else if (index == SAL_CLEAR_STATE_INFO) {
		;
	} else if (index == SAL_MC_RENDEZ) {
		;
	} else if (index == SAL_MC_SET_PARAMS) {
		;
	} else if (index == SAL_CACHE_FLUSH) {
		;
	} else if (index == SAL_CACHE_INIT) {
		;
#ifdef CONFIG_PCI
	} else if (index == SAL_PCI_CONFIG_READ) {
		/*
		 * in1 contains the PCI configuration address and in2
		 * the size of the read.  The value that is read is
		 * returned via the general register r9.
		 */
		outl(BUILD_CMD(in1), 0xCF8);
		if (in2 == 1)	/* Reading byte  */
			r9 = inb(0xCFC + ((REG_OFFSET(in1) & 3)));
		else if (in2 == 2)	/* Reading word  */
			r9 = inw(0xCFC + ((REG_OFFSET(in1) & 2)));
		else		/* Reading dword */
			r9 = inl(0xCFC);
		status = PCIBIOS_SUCCESSFUL;
	} else if (index == SAL_PCI_CONFIG_WRITE) {
		/*
		 * in1 contains the PCI configuration address, in2 the
		 * size of the write, and in3 the actual value to be
		 * written out.
		 */
		outl(BUILD_CMD(in1), 0xCF8);
		if (in2 == 1)	/* Writing byte  */
			outb(in3, 0xCFC + ((REG_OFFSET(in1) & 3)));
		else if (in2 == 2)	/* Writing word  */
			outw(in3, 0xCFC + ((REG_OFFSET(in1) & 2)));
		else		/* Writing dword */
			outl(in3, 0xCFC);
		status = PCIBIOS_SUCCESSFUL;
#endif				/* CONFIG_PCI */
	} else if (index == SAL_UPDATE_PAL) {
		;
	} else {
		status = -1;
	}
	return ((struct sal_ret_values) {
		status, r9, r10, r11});
}


static int num_mem_descs = 0;

static efi_memory_desc_t *make_md(u32 typ, u64 attr, u64 start, u64 end)
{
	efi_memory_desc_t *md = bloballoc(sizeof(efi_memory_desc_t));
	md->type = typ;
	md->pad = 0;
	md->phys_addr = start;
	md->virt_addr = 0;
	md->num_pages = (end - start) >> 12;
	md->attribute = attr;
	num_mem_descs++;

	return md;
}

static long env_or_constant(char *var, long fallback)
{
	char *val = getenv(var);

	if (val)
		return atoi(val);

	return fallback;
}

void *get_bootparams(unsigned long lowmem, unsigned long highmem,
		     char *cmd_line)
{
	ia64_boot_param_t *bp = bloballoc(sizeof(ia64_boot_param_t));
	char *args = bloballoc(256);
	efi_system_table_t *efi_systab =
	    bloballoc(sizeof(efi_system_table_t));
	efi_runtime_services_t *efi_runtime =
	    bloballoc(sizeof(efi_runtime_services_t));
	efi_config_table_t *efi_tables =
	    bloballoc(sizeof(efi_config_table_t));
	struct ia64_sal_systab *sal_systab =
	    bloballoc(sizeof(struct ia64_sal_systab));
	struct ia64_sal_desc_entry_point *sal_ed =
	    bloballoc(sizeof(struct ia64_sal_desc_entry_point));
	efi_memory_desc_t *efi_memmap, *md;
	unsigned long *pal_desc, *sal_desc;
	unsigned char checksum = 0;
	int i;

	pal_desc = (unsigned long *) &pal_emulator_static;
	sal_desc = (unsigned long *) &sal_emulator;

	efi_systab->hdr.signature = EFI_SYSTEM_TABLE_SIGNATURE;
	efi_systab->hdr.revision = EFI_SYSTEM_TABLE_REVISION;
	efi_systab->hdr.headersize = sizeof efi_systab->hdr;
	efi_systab->fw_vendor = __pa("G\0e\0l\0a\0t\0o\0\0\0");
	efi_systab->fw_revision = 1;
	efi_systab->runtime = (void *) __pa(efi_runtime);
	efi_systab->nr_tables = 1;
	efi_systab->tables = __pa(efi_tables);

	efi_runtime->hdr.signature = EFI_RUNTIME_SERVICES_SIGNATURE;
	efi_runtime->hdr.revision = EFI_RUNTIME_SERVICES_REVISION;
	efi_runtime->hdr.headersize = sizeof(efi_runtime->hdr);
	efi_runtime->get_time = __pa(&fw_efi_get_time);
	efi_runtime->set_time = __pa(&efi_unimplemented);
	efi_runtime->get_wakeup_time = __pa(&efi_unimplemented);
	efi_runtime->set_wakeup_time = __pa(&efi_unimplemented);
	efi_runtime->set_virtual_address_map = __pa(&efi_unimplemented);
	efi_runtime->get_variable = __pa(&efi_unimplemented);
	efi_runtime->get_next_variable = __pa(&efi_unimplemented);
	efi_runtime->set_variable = __pa(&efi_unimplemented);
	efi_runtime->get_next_high_mono_count = __pa(&efi_unimplemented);
	efi_runtime->reset_system = __pa(&efi_reset_system);

	efi_tables->guid = SAL_SYSTEM_TABLE_GUID;
	efi_tables->table = __pa(sal_systab);

	/* fill in the SAL system table: */
	memcpy(sal_systab->signature, "SST_", 4);
	sal_systab->size = sizeof(*sal_systab);
	sal_systab->sal_rev_minor = 1;
	sal_systab->sal_rev_major = 0;
	sal_systab->entry_count = 1;

	strcpy((char *) sal_systab->oem_id, "Generic");
	strcpy((char *) sal_systab->product_id, "IA-64 system");

	/* fill in an entry point: */
	sal_ed->type = SAL_DESC_ENTRY_POINT;
	sal_ed->pal_proc = __pa(pal_desc[0]);
	sal_ed->sal_proc = __pa(sal_desc[0]);
	sal_ed->gp = __pa(sal_desc[1]);

	for (i = 0; i < sizeof sal_systab; i++)
		checksum += ((char *) sal_systab)[i];

	sal_systab->checksum = -checksum;

	/* simulate free memory at physical address zero */
	efi_memmap =
	    make_md(EFI_BOOT_SERVICES_DATA, EFI_MEMORY_WB, (512 * 1024),
		    1 * MB);
	make_md(EFI_PAL_CODE, EFI_MEMORY_WB, 1 * MB, 2 * MB);
	make_md(EFI_CONVENTIONAL_MEMORY, EFI_MEMORY_WB, 2 * MB, lowmem);
	make_md(EFI_LOADER_DATA, EFI_MEMORY_WB, lowmem,
		highmem + ((1 << 12) - 1));
	md = make_md(EFI_CONVENTIONAL_MEMORY, EFI_MEMORY_WB,
		     (((highmem + ((1 << 12) - 1)) >> 12) + 1) << 12,
		     0 * MB);
	md->num_pages = get_memsize((unsigned long *) &md->phys_addr);

	for (i = 0; i < num_mem_descs; i++)
		fprintf(stderr,
			"efimem addr:%p t:%d a:%lx p:%lx v:%lx pages:%lx\n",
			&efi_memmap[i], efi_memmap[i].type,
			efi_memmap[i].attribute, efi_memmap[i].phys_addr,
			efi_memmap[i].virt_addr, efi_memmap[i].num_pages);
	bp->efi_systab = __pa(efi_systab);
	bp->efi_memmap = __pa(efi_memmap);
	bp->efi_memmap_size = num_mem_descs * sizeof(efi_memory_desc_t);
	bp->efi_memdesc_size = sizeof(efi_memory_desc_t);
	bp->efi_memdesc_version = 1;
	strcpy(args, cmd_line);
	bp->command_line = __pa(args);
	fprintf(stderr, "bp->command_line @%p\n", args);
	bp->console_info.num_cols = env_or_constant("COLUMNS", 80);
	bp->console_info.num_rows = env_or_constant("ROWS", 40);
	bp->console_info.orig_x = 0;
	bp->console_info.orig_y = 24;
	bp->fpswa = 0;

	return bp;
}
