/* emu.c: Emulation of privileged instructions */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <asm/rse.h>
#include "vm.h"
#include "debug.h"


#define M (0 << 4)
#define I (1 << 4)
#define F (2 << 4)
#define L (3 << 4)
#define B (4 << 4)
#define R (5 << 4)

static const char templates[16][4] = {
	{ M, I, I }, { M, I, I }, { M, L, R }, { R, R, R },
	{ M, M, I }, { M, M, I }, { M, F, I }, { M, M, F },
	{ M, I, B }, { M, B, B }, { R, R, R }, { B, B, B },
	{ M, M, B }, { R, R, R }, { M, F, B }, { R, R, R }
};

extern unsigned long smp_rendezvous_ip, smp_rendezvous_gp;


struct cpu cpu;

static inline void
load_scratch(struct sigcontext *sc)
{
	asm volatile (
		"mov b2=%0\n\t"
		"mov r5=%1\n\t"
		"mov r6=%2\n\t"
		"mov r7=%3\n\t"
		"mov r4=%4\n\t"
		::
		 "r"(cpu.saved_b2),
		 "r"(cpu.gr[1]),
		 "r"(cpu.gr[2]),
		 "r"(cpu.gr[3]),
		 "r"(cpu.gr[0]));
}

unsigned long
read_register(struct sigcontext *sc, unsigned int regno)
{
	unsigned long *bsp;
	if (regno == 0)
	return 0;
	if (regno < 32)
	{
		if ((regno >= 4) && (regno <= 7))
		{
			unsigned long r;
			asm volatile  ("mov %0=r4" : "=r"(r));
			if (r) {
				switch (regno) {
				case 4:
					break;
				case 5:
					asm volatile ("mov %0=r5" : "=r"(r));
					break;
				case 6:
					asm volatile ("mov %0=r6" : "=r"(r));
					break;
				case 7:
					asm volatile ("mov %0=r7" : "=r"(r));
					break;
				}
				return r;
			}
			if (regno != 4)
				vm_panic(sc, "[%lx] bad register %d\n", sc->sc_ip, regno);
			return r;
		}

		return sc->sc_gr[regno];
	}
	if (BITS(sc->sc_cfm, 14, 4) != 0)
		abort();
	asm volatile(";; flushrs ;;");

	/* see comment in asm/sigcontext.h re alternate stack */
	/* bsp = (unsigned long *)(sc->sc_ar_bsp); */
	if (sc->sc_rbs_base == 0)
		abort();
	bsp = (unsigned long *)(sc->sc_rbs_base + (sc->sc_loadrs >> 16));
	bsp = ia64_rse_skip_regs(bsp, -(sc->sc_cfm & 0x7f) + (regno - 32));
	if ((unsigned long)bsp < sc->sc_rbs_base)
		abort();
	if (ia64_rse_is_rnat_slot(bsp))
		abort();
	return *bsp;
}

void
write_register(struct sigcontext *sc, unsigned int regno, unsigned long value)
{
	unsigned long *bsp;

	if (regno < 32)
	{
		if ((regno == 0) || ((regno >= 4) && (regno <= 7)))
		{
			unsigned long r;
			asm volatile  ("mov %0=r4" : "=r"(r));
			if (regno && r) {
				switch(regno) {
				case 5:
					asm volatile ("mov r5=%0" : :"r"(value));
					break;
				case 6:
					asm volatile ("mov r6=%0" : :"r"(value));
					break;

				case 7:
					asm volatile ("mov r7=%0" : :"r"(value));
					break;
				default:
					goto bad;
				}
				return;

			}
		bad:
			fprintf(stderr, "[%lx] write bad register %d\n", sc->sc_ip, regno);
			abort();
		}
		sc->sc_gr[regno] = value;
		sc->sc_nat &= ~(1UL << regno);
	}
	else
	{
		if (BITS(sc->sc_cfm, 14, 4) != 0)
			abort();
		asm volatile(";; mov ar.rsc = 0 ;; flushrs ;; loadrs ;;");
		/* see comment in asm/sigcontext.h re alternate stack */
		/* bsp = (unsigned long *)(sc->sc_ar_bsp); */
		if (sc->sc_rbs_base == 0)
			abort();
		bsp = (unsigned long *)(sc->sc_rbs_base + (sc->sc_loadrs >> 16));
		bsp = ia64_rse_skip_regs(bsp, -(sc->sc_cfm & 0x7f) + (regno - 32));
		if ((unsigned long)bsp < sc->sc_rbs_base)
			abort();
		if (ia64_rse_is_rnat_slot(bsp))
			abort();
		*bsp = value;
		/* FIXME: clear NaT */
		asm volatile(";; mov ar.rsc = 3 ;;");
	}
}

#define PTRACE_FPH_TRAP 0x26
unsigned long inline
read_psr(struct sigcontext *sc)
{
	return cpu.psr | ((sc->sc_um | (PSR_MFH | PSR_MFL)) & 0x3fUL);
}

void
write_psr(struct sigcontext *sc, unsigned long value)
{
	unsigned int bank;
	static int can_ptrace_dfh = 1;


	bank = BITS(value, PSR_BN_BIT, 1);
	emu_bsw(sc, bank);
	cpu.psr = value;

	/*
	 * This relies on a hacked 
	 * host kernel.
	 */
	if (BITS(cpu.psr, PSR_DFH_BIT, 1) && can_ptrace_dfh)
	{
		if (ptrace(PTRACE_FPH_TRAP) == -1) {
			fprintf(stderr, "Can't ptrace_FPH_trap\n");
			can_ptrace_dfh = 0;
		}
	}
	sc->sc_um = cpu.psr & 0x3f;
	/*
	 * going to user space, so 
	 * reload r4--r7 and b2.
	 * Do this *before* check_interrupts, so that
	 * any exception handler gets the right values.
	 */
	if (cpu.psr & PSR_CPL)
		load_scratch(sc);
	check_interrupts(sc);
}

/*
 * Local memcpy, because GCC builtin/glibc one uses the FP registers
 */
static inline void my_memcpy(unsigned long *dst, const unsigned long *src, size_t n)
{
	n /= sizeof (long);
	while (n-- > 0)
		*dst++ = *src++;
}

void
emu_bsw(struct sigcontext *sc, unsigned int target_bank)
{
	if (!!(cpu.psr & PSR_BN) == target_bank)
		return;
	DEBUG_EMU("BANK SWITCH: %d taken\n", target_bank);

	my_memcpy(&cpu.saved_bank[16*(1-target_bank)], &sc->sc_gr[16], 16*sizeof (long));
	my_memcpy(&sc->sc_gr[16], &cpu.saved_bank[16*target_bank], 16*sizeof (long));

	cpu.psr ^= PSR_BN;
}

void
emu_cover(struct sigcontext *sc)
{
	if (cpu.cfle)
		cpu.cr[CR_IFS] = sc->sc_cfm | CFM_VALID;
	else
		cpu.cfle = 1;

	sc->sc_cfm = 0;
}

void
emu_tpa(struct sigcontext *sc, int src, int dst)
{
	unsigned long addr = read_register(sc, src);
	unsigned long value;

	/*
	 * TPA faults if not executed in privileged mode
	 */
#if 0
	if (cpu.psr & PSR_CPL) {
		deliver_exception(sc, GENERAL_EXCEPTION, 0);
		return;
	}
#endif
	value = tpa(addr);
	if (value)
		write_register(sc, dst, value);
	else {
		sc->sc_ip -= ((sc->sc_ip & 0xf) == 0) ? 0xe : 1;
		DEBUG_EMU("tpa faulted %lx %lx\n", sc->sc_ip, addr);
		mm_fault(sc, addr, 0, 0);
	}
}

/*
 * with afterburning this should never be invoked
 */
void
emu_M0(struct sigcontext *sc, unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;
	unsigned long mask = BITS(instr, 6, 21) | ((x6 >> 4) << 21) | (BITS(instr, 36, 1) << 23);

	switch (x6)
	{
		case 0:
			/* BREAK */
			fprintf(stderr, "[%lx] BREAK (emu) %lx\n", sc->sc_ip, (((instr>>36) & 1)<<20)| ((instr>>6) & 0xfffff) );
			/*
			 * For Linux this is a BUG() statement: just bail out.
			 */
#if 0
			if ((sc->sc_ip > (4UL<<61)) && ((((instr>>36) & 1)<<20)|((instr>>6) & 0xfffff)) == 0)
				vm_panic(sc, "Kernel BUG()\n");
#endif
			cpu.cr[CR_IIM] = 0;
			deliver_exception(sc, BREAK_EXCEPTION, 0);
			break;

		case 0x6:
		case 0x16:
		case 0x26:
		case 0x36:
			DEBUG_EMU("[%lx] SSM 0x%lx\n", sc->sc_ip, mask);
			INC_STAT(ssm);
			write_psr(sc, read_psr(sc) | mask);
			break;

		case 0x7:
		case 0x17:
		case 0x27:
		case 0x37:
			DEBUG_EMU("[%lx] RSM 0x%lx\n", sc->sc_ip, mask);
			INC_STAT(rsm);
			write_psr(sc, read_psr(sc) & ~mask);
			break;

		default:
			fprintf(stderr, "%lx: M0: %lx %lx\n", sc->sc_ip, x6, sc->sc_br[0]);
			abort();
	}

}

unsigned long itcval = 0;

void tlbop(struct sigcontext *sc, int code, unsigned int r0, unsigned int r1)
{

	DEBUG_EMU("tlbop(%d, r%d = %lx , r%d=%lx)\n",
		code, r0, read_register(sc, r0), r1, read_register(sc, r1));

	switch (code)
	{
	case 0: /* itc.d */
	case 1: /* itc.i */
		if (tlb_insert(read_register(sc, r0), 0))
			vm_panic(sc, "tlbop01 failed\n");
		break;
	case 2:			/* itr.d */
	case 3:			/* itr.i */
		if (tlb_insert(read_register(sc, r0), 1))
			vm_panic(sc, "tlbop23 failed\n");
		break;
	
	case 4: /* ptc.l */
		tlb_purge(read_register(sc, r0), read_register(sc, r1));
		break;

	case 5: /* ptc.g */
	case 6: /* ptc.ga */
		tlb_purge_global(sc, read_register(sc, r0), read_register(sc, r1));
		break;
		
	case 7: /* ptr.d */
	case 8: /* ptr.i */
		fprintf(stderr, "ptr.[id] Not yet implemented\n");
		break;

	case 9: /* ptc.e */
		tlb_purge_all();
		break;

	default: 
		fprintf(stderr, "Bad tlbop %d\n", code);
		abort();
	}
}

void
emu_write_rr(struct sigcontext *sc, int rgnreg, int valuereg)
{
	int region;
	unsigned long value;

	region = read_register(sc, rgnreg) >> 61;
	value = read_register(sc, valuereg);
	DEBUG_EMU("[%lx] MOVE to RR %d = %lx\n", sc->sc_ip, region, value);
	if (BITS(value, 8, 24) != BITS(cpu.rr[region], 8, 24))
	{
		/* RID changed */
#ifdef RR_CHANGE_ALWAYS_PURGE
		tlb_purge_region(region);
#else
		set_asid(BITS(value, 8, 24), region);
#endif
#if (DEBUG)&(DEBUG_MM_BIT)
		if (region == 3)
			fprintf(stderr, "context switch from %lx to %lx\n", BITS(cpu.rr[region], 8, 24),
			       BITS(value, 8, 24));
#endif
	}
	cpu.rr[region] = value;
}


void
emu_M1(struct sigcontext *sc, unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;
	unsigned int r1 = (instr >> 6) & 0x7f;
	unsigned int r2 = (instr >> 13) & 0x7f;
	unsigned int r3 = (instr >> 20) & 0x7f;
	unsigned int region;

	switch (x6)
	{
		case 0x00:	
			INC_STAT(mov_to_rr);
			emu_write_rr(sc, r3, r2);
			break;

		case 0x03:	
			INC_STAT(mov_to_pkr);
			/* MOVE to PKR */
			break;

		case 0x09: /* PTC.L */
			DEBUG_EMU("[%lx] PTC.L\n", sc->sc_ip);
			INC_STAT(ptc_l);
			tlbop(sc, 4, r3, r2);
			break;

		case 0x0a: /* PTC.G */
		case 0x0b: /* PTC.GA */
			DEBUG_EMU("[%lx] PTC.G\n", sc->sc_ip);	
			INC_STAT(ptc_g);
			tlbop(sc, 5, r3, r2);
			break;

		case 0x0e:
			DEBUG_EMU("[%lx] ITR.D\n", sc->sc_ip);
			INC_STAT(itr);
			tlbop(sc, 1, r2, 0);
			break;

		case 0x0f:
			DEBUG_EMU("[%lx] ITR.I\n", sc->sc_ip);
			INC_STAT(itr);
			tlbop(sc, 2, r2, 0);
			break;

		case 0x10:
			INC_STAT(mov_from_rr);
			region = read_register(sc, r3) >> 61;
			DEBUG_EMU("[%lx] MOVE from RR %d\n", sc->sc_ip, region);
			write_register(sc, r1, cpu.rr[region]);
			break;

		case 0x14: /* MOVE from PMC */
			DEBUG_EMU("[%lx] THASH\n", sc->sc_ip);
			INC_STAT(mov_from_pmc);
			write_register(sc, r1, thash(read_register(sc, r3)));
			break;

		case 0x15: /* MOVE from PMD */
			DEBUG_EMU("[%lx] TTAG\n", sc->sc_ip);	
			INC_STAT(mov_from_pmd);
			write_register(sc, r1, ttag(read_register(sc, r3)));
			break;

		case 0x1e: /* TPA */
			DEBUG_EMU("[%lx] TPA\n", sc->sc_ip);
			INC_STAT(ntpa);
			emu_tpa(sc, r3, r1);
			break;

#if 0
		case 0x22:
			itcval += 0x800000;
			DEBUG_EMU("[%lx] MOVE from AR %d = %lx\n", sc->sc_ip, r3, itcval);
			//assert(r3 = 44);
			write_register(sc, r1, itcval);
			break;
#endif

		case 0x24:
			DEBUG_EMU("[%lx] MOVE from CR %d = %lx\n", sc->sc_ip, r3, cpu.cr[r3]);
			INC_STAT(mov_from_cr);
			write_register(sc, r1, cpu.cr[r3]);
			switch (r3)
			{
				case CR_IVR:
					cpu.in_service_irq = cpu.cr[CR_IVR];
					clear_interrupt(cpu.cr[CR_IVR]);
					break;
			}
			break;

		case 0x25:
			DEBUG_EMU("[%lx] MOVE from PSR = %lx\n", sc->sc_ip, read_psr(sc));	
			INC_STAT(nread_psr);
			write_register(sc, r1, read_psr(sc));
			break;

		case 0x2a:
			DEBUG_EMU("[%lx] MOVE to AR %d\n", sc->sc_ip, r3);
			INC_STAT(mov_to_ar);

			if (r3 < 8)
			{
				cpu.kr[r3] = read_register(sc, r2);
			}
			else if (r3 == 44) /* ar.itc */
			{
			}
			else if (r3 == 16) {
				asm volatile(";; mov ar.rsc = 3 ;;");
			}
			else
			{
				fprintf(stderr, "[%lx] move to bad AR %d\n", sc->sc_ip, r3);
				fprintf(stderr, "Value = %lx\n", read_register(sc, r2));
				vm_panic(NULL, "Bailing out");
			}
			break;

		case 0x2c:
			cpu.cr[r3] = read_register(sc, r2);
			INC_STAT(mov_to_cr);

			DEBUG_EMU("[%lx] MOVE to CR %d = 0x%lx\n", sc->sc_ip, r3, cpu.cr[r3]);
			switch (r3)
			{
				case CR_EOI:
					cpu.in_service_irq = 0;
					break;
				case CR_ITM:
					set_timer();
					break;
			}
			break;

		case 0x2d:
			DEBUG_EMU("[%lx] MOVE to PSR.l\n", sc->sc_ip);
			INC_STAT(mov_to_psr);

			write_psr(sc, (read_register(sc, r2) & 0xffffffffUL) | (read_psr(sc) >> 32 << 32));
			break;

		case 0x2e:
			DEBUG_EMU("[%lx] ITC.D\n", sc->sc_ip);
			INC_STAT(itc_d);
			if (tlb_insert(read_register(sc, r2), 0))
				vm_panic(sc, "itc.d failed\n");
			break;

		case 0x2f:
			DEBUG_EMU("[%lx] ITC.I\n", sc->sc_ip);
			INC_STAT(itc_i);
			if (tlb_insert(read_register(sc, r2), 0))
				vm_panic(sc, "itc.i failed\n");
			break;

		case 0x34:
			DEBUG_EMU("[%lx] PTC.E\n", sc->sc_ip);
			INC_STAT(ptc_e);
			tlb_purge_all();
			break;

		default:
			fprintf(stderr, "[%lx] M1: %lx\n", sc->sc_ip, x6);
			abort();
	}
}

void
emu_B0(struct sigcontext *sc, unsigned long instr)
{
	unsigned long x6 = (instr >> 27) & 0x3f;

	switch (x6)
	{
		case 0x8:
			DEBUG_EMU("[%lx] RFI -> %lx\n", sc->sc_ip, cpu.cr[CR_IIP]);
	
			INC_STAT(rfi);
			if ((cpu.cr[CR_IFS] & CFM_VALID) && ((sc->sc_cfm & 0x7f) != 0))
			{
				extern void ShrinkFrame;
				INC_STAT(shrinkframes);
				sc->sc_ip = (unsigned long)&ShrinkFrame;
				/* fprintf(stderr, "ShrinkFrame\n");*/
				break;
			}

			sc->sc_ip = (cpu.cr[CR_IIP] & ~0xf) | (BITS(cpu.cr[CR_IPSR], PSR_RI_OFFSET, 2));
			if (cpu.cr[CR_IFS] & CFM_VALID)
				sc->sc_cfm = BITS(cpu.cr[CR_IFS], 0, 38);
			cpu.cfle = 1;
			write_psr(sc, cpu.cr[CR_IPSR] & ~PSR_RI);
			break;

		case 0xc:
		case 0xd:
	
			INC_STAT(bsw);
			emu_bsw(sc, x6 & 1);
			break;

		default:
			fprintf(stderr, "%lx: emu_B0: %lx\n", sc->sc_ip, x6);
			abort();
	}
}

/* emulate a privileged instruction */
void
emu_priv(struct sigcontext *sc)
{
	unsigned long ip = sc->sc_ip;
	unsigned int slot = ip & 0xf;
	unsigned long w1, w2, instr;
	unsigned int type, opcode;

	sc->sc_ip += (slot == 2) ? 0xe : 1;

	ip &= ~0xfL;
	w1 = *(unsigned long *)ip;
	w2 = *(unsigned long *)(ip+8);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			break;
		case 2:
			instr = (w2 >> 23);
			break;
		default:
			fprintf(stderr, "Bad slot %d\n", slot);
			abort();
			instr = 0;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 37) & 0xf;

	if ((type == M) && (opcode == 0))
		emu_M0(sc, instr);
	else if ((type == M) && (opcode == 1))
		emu_M1(sc, instr);
	else if ((type == B) && (opcode == 0))
		emu_B0(sc, instr);
	else {
      		sc->sc_ip -= (slot == 2) ? 0xe : 1;
		if (cpu.psr & PSR_CPL) /* userspace */ 
			deliver_exception(sc, GENERAL_EXCEPTION, 0);
		else
			vm_panic(sc, "emu_priv failed: ip %lx, type %x, opcode %x\n", sc->sc_ip, type, opcode);
	}
}

#ifdef SMP
/* emulate a store to the Processor Interrupt Block */
void
emu_store_PIB(struct sigcontext *sc, unsigned long addr)
{
	unsigned long ip = sc->sc_ip;
	unsigned int slot = ip & 0xf;
	unsigned long w1, w2, instr;
	unsigned int type, opcode, r2, r3, target;
	unsigned long staddr, value;
	struct request request;

	sc->sc_ip += (slot == 2) ? 0xe : 1;

	ip &= ~0xf;
	w1 = *(unsigned long *)ip;
	w2 = *(unsigned long *)(ip+8);

	switch (slot)
	{
		case 0:
			instr = (w1 >> 5);
			break;
		case 1:
			instr = (w1 >> 46) | (w2 << 18);
			break;
		case 2:
			instr = (w2 >> 23);
			break;
		default:
			fprintf(stderr, "Bad slot %d\n", slot);
			abort();
			instr = 0;
	}

	type = templates[(w1 >> 1) & 0xf][slot];
	opcode = (instr >> 37) & 0xf;

	if ((type != M) || (opcode != 4))
	{
		fprintf(stderr, "emu_store_PIB: ip %lx, type %x, opcode %x\n", sc->sc_ip, type, opcode);
		return;
	}

	r2 = (instr >> 13) & 0x7f;
	r3 = (instr >> 20) & 0x7f;
	staddr = read_register(sc, r3);
	value = read_register(sc, r2);
	if (staddr != addr)
	{
		fprintf(stderr, "emu_store_PIB: staddr %lx, addr %lx\n", staddr, addr);
		return;
	}

	target = (addr >> 4) & 63;

	if (value == 0xff)
	{
		request.type = IPC_STARTUP;
		request.startup.ip = smp_rendezvous_ip;
		request.startup.gp = smp_rendezvous_gp;
	}
	else
	{
		request.type = IPC_IPI;
		request.ipi.irq = value;
	}

	ipc_send(target, &request);
}
#endif
