/* exception.c: exceptions and interrupts */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include "vm.h"
#include "debug.h"

static inline void
save_scratch(void)
{

	asm ("st8 [%1]=r4; st8 [%2]=r5; st8 [%3]=r6; st8 [%4]=r7; mov %0=b2; ;; ; mov r4=r0" 
	     :  "=r"(cpu.saved_b2) :
	     "r"(&cpu.gr[0]), "r"(&cpu.gr[1]), "r"(&cpu.gr[2]), 
	     "r"(&cpu.gr[3]));
}


void
deliver_exception(struct sigcontext *sc, unsigned long vector, unsigned long isr)
{
	extern void _text, _etext, ShrinkFrame;
	unsigned long ri;

	/*
	 * Save if we were running unprivileged
	 */
	if (cpu.psr & PSR_CPL)
		save_scratch();
	
	if (isr & (1UL << 38)) /* incomplete register frame */
	{
		DEBUG_EXCEPTION("incomplete register frame\n"); 
		cpu.cr[CR_IFS] = sc->sc_cfm | (1UL << 63);
		sc->sc_cfm = 0;
		cpu.cfle = 0;
		if (!(cpu.psr & PSR_IC))
			return;
	} else if (cpu.psr & PSR_IC)
		cpu.cr[CR_IFS] = 0;

	if (cpu.psr & PSR_IC) {
 		if (vector == 0x2c00) /* break */
 			sc->sc_ip -= ((sc->sc_ip & 0x3) == 0) ? 0xe : 1;
		ri = (sc->sc_ip & 0x3) << 41;
		cpu.cr[CR_IIP] = sc->sc_ip & ~0xf;
		DEBUG_EXCEPTION("Exception %lx at %lx\n", vector, cpu.cr[CR_IIP]);
		cpu.cr[CR_IIPA] = cpu.cr[CR_IIP];
		cpu.cr[CR_ISR] = (isr & ~PSR_RI) | ri;
		cpu.cr[CR_IPSR] = (cpu.psr & ~(PSR_RI | 0x3f)) | ri | (sc->sc_um & 0x3f) | PSR_MFH;
#ifdef AFTERBURN
		/* Afterburner hack */
		if (!(cpu.psr & PSR_CPL)) {
			unsigned long restart_ip;

			asm volatile  ("mov %0=r4" : "=r"(restart_ip));
			if (restart_ip) {
				DEBUG_EXCEPTION("Rewriting restart IP to %lx\n",
					restart_ip);

				cpu.cr[CR_IIP] = restart_ip;
				cpu.cr[CR_IPSR] &= ~PSR_RI;
			}
		}
#endif
	}
	else if (vector != 0x1400) {
			vm_panic(sc, "Exception %lx with psr.ic off: %lx\n", vector, sc->sc_ip);
	}
	if ((sc->sc_ip >= (unsigned long)&_text) 
	    && (sc->sc_ip < (unsigned long)&_etext) 
	    && ((sc->sc_ip & ~0xfUL) != (unsigned long)&ShrinkFrame)) {
		vm_panic(sc, "oops: deliver_exception inside VM (%lx)?!\n", sc->sc_ip);
	}


	if (cpu.cr[CR_IVA] == 0) {
		dump_maps();
		vm_panic(NULL, "exception with cr_iva==0");
	}

	DEBUG_EXCEPTION("Deliver_exception %lx\n", vector);
	sc->sc_ip = cpu.cr[CR_IVA] + vector;
	emu_bsw(sc, 0);
	cpu.psr &= ~(PSR_BN | PSR_CPL | PSR_I | PSR_IC);
}

void
check_interrupts(struct sigcontext *sc)
{
	if (cpu.irq_pending && (cpu.psr & PSR_I)
	    && !(cpu.cr[CR_TPR] & TPR_MMI)
	    && (cpu.cr[CR_IVR] > cpu.in_service_irq)
	    && ((cpu.cr[CR_IVR] >> 4) > (cpu.cr[CR_TPR] >> 4)))
	{
		deliver_exception(sc, EXT_IRQ_VECTOR, 0);
	}
}

void
deliver_interrupt(struct sigcontext *sc, unsigned int irq)
{

	cpu.cr[CR_IRR0+(irq/64)] |= (1UL << (irq % 64));
	if (irq > cpu.cr[CR_IVR])
		cpu.cr[CR_IVR] = irq;
	cpu.irq_pending = 1;
	check_interrupts(sc);
}

static unsigned long
ia64_fls(unsigned long x)
{
	long double d = x;
	long exp;

	__asm__ ("getf.exp %0=%1" : "=r"(exp) : "f"(d));
	return exp - 0xffff;
}

void
clear_interrupt(unsigned int irq)
{
	int i;

	cpu.cr[CR_IRR0+(irq/64)] &= ~(1UL << (irq % 64));
	for (i = 3; i >= 0; i--)
	{
		if (cpu.cr[CR_IRR0+i] != 0)
		{
			cpu.cr[CR_IVR] = ia64_fls(cpu.cr[CR_IRR0+i]) + (i*64);
			return;
		}
	}

	cpu.cr[CR_IVR] = 0xf;
	cpu.irq_pending = 0;
}

