#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include "vm.h"
#include "debug.h"

extern unsigned long smp_rendezvous_ip, smp_rendezvous_gp;
extern unsigned long pagesize;

extern int cpus;
int sock[MAXCPU];
int ready;

#ifdef DEBUG_IPC
static char *messages[] = {
	"IPC_FETCH",
	"IPC_FETCH_REPLY",
	"IPC_FETCH_NAK",
	"IPC_INVALIDATE",
	"IPC_INVALIDATE_REPLY",
	"IPC_PURGE",
	"IPC_PURGE_REPLY",
	"IPC_IPI",
	"IPC_STARTUP",
	"IPC_SSC",
	"IPC_SSC_REPLY",
	"IPC_SHUTDOWN",
};
#endif

void ipc_init(void)
{
	struct sockaddr_in servaddr;
	int listensock, i;
	int true_value = 1;

	listensock = socket(AF_INET, SOCK_STREAM, 0);
	setsockopt(listensock, SOL_SOCKET, SO_REUSEADDR, (void *) &true_value, sizeof(true_value));

	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(VM_PORT_BASE+cpu.id);
	servaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
	if (bind(listensock, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
	{
		vm_perror("bind");
		return;
	}

	listen(listensock, cpu.id);

	for (i = 0; i < cpu.id; i++)
	{
		int newsock = accept(listensock, NULL, 0);
		unsigned char sockindex;
		read (newsock, &sockindex, 1);
		sock[sockindex] = newsock;
		if (sock[i] == -1)
		{
			perror("accept");
			return;
		}
		IPC_DEBUG("Fd %d to cpu %d\n", sock[i], i);
	}

	for (i = cpu.id+1; i < cpus; i++)
	{
		sock[i] = socket(AF_INET, SOCK_STREAM, 0);
		servaddr.sin_port = htons(VM_PORT_BASE+i);
		while (connect(sock[i], (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
		{
			if (errno == ECONNREFUSED)
			{
				sleep(1);
				continue;
			}
			perror("connect");
			return;
		}
 		char buf = cpu.id; 
 		write(sock[i], &buf, 1);
		IPC_DEBUG("Fd %d to cpu %d\n", sock[i], i);
	}

	close(listensock);

	for (i = 0; i < cpus; i++)
	{
		if (i == cpu.id)
			continue;

		setsockopt(sock[i], IPPROTO_TCP, TCP_NODELAY, (void *) &true_value, sizeof(true_value));
		fcntl(sock[i], F_SETOWN, getpid());
		fcntl(sock[i], F_SETSIG, SIGIPC);
		fcntl(sock[i], F_SETFL, fcntl(sock[i], F_GETFL)|O_ASYNC);
	}

	if (cpu.id != 0)
	{
		sigset_t sigset;

		/* wait for IPC_STARTUP */
		sigemptyset(&sigset);
		sigaddset(&sigset, SIGIPC);
		sigprocmask(SIG_BLOCK, &sigset, NULL);
		ipc_recv(NULL);
		sigprocmask(SIG_UNBLOCK, &sigset, NULL);
	}

	ready = 1;
}

void
ipc_send_reply(int node, unsigned char type)
{
	struct request request;

	request.type = type;
	ipc_send(node, &request);
}

void
ipc_send_fetch_reply(int node, unsigned long pa, int is_write)
{
	struct request request;

	request.type = IPC_FETCH_REPLY;
	request.fetch.pa = pa;
	request.fetch.is_write = is_write;
	ipc_send(node, &request);

	if (send(sock[node], (void *)(request.fetch.pa), pagesize, 0) != pagesize)
		vm_perror("fetch send");
}

int
ipc_handle_async(struct sigcontext *sc, int fd)
{
	struct request request;
	struct {
		unsigned long addr;
		unsigned len;
	} disk_req;
	int i, node = -1, ret = 0;

	for (i = 0; i < cpus; i++)
	{
		if (sock[i] == fd)
			node = i;
	}

	if (node == -1)
		return 0;

	IPC_DEBUG("ipc_handle_async\n");
#ifdef DSM
	/* process queued requests first */
	process_request_queue(sc);
#endif

	while (recv(fd, &request, sizeof(request), MSG_WAITALL|MSG_DONTWAIT) == sizeof(request))
	{
		IPC_DEBUG("recv message %s\n", messages[request.type]);

		switch (request.type)
		{
#ifdef DSM
		case IPC_FETCH:
			mm_handle_fetch(sc, node, request.fetch.pa, request.fetch.is_write);
			break;
		case IPC_FETCH_REPLY:
			if (mprotect((void *)(request.fetch.pa), pagesize, PROT_READ|PROT_EXEC|PROT_WRITE))
				vm_perror("mprotect");
			if (recv(sock[node], (void *)(request.fetch.pa), pagesize, MSG_WAITALL) != pagesize)
				vm_perror("response recv");
			mm_handle_fetch_reply(request.fetch.pa, request.fetch.is_write);
			ret = 1;
			break;
		case IPC_FETCH_NAK:
			mm_handle_fetch_nak(node, request.fetch.pa, request.fetch.is_write);
			break;
		case IPC_INVALIDATE:
			mm_handle_invalidate(node, request.invalidate.pa, request.invalidate.is_write);
			break;
		case IPC_INVALIDATE_REPLY:
			mm_handle_invalidate_reply(request.invalidate.pa, request.invalidate.is_write);
			ret = 1;
			break;
#endif
		case IPC_PURGE:
			/* HACK: if CPU not "ready", then our address space layout is different */
			if (ready)
				tlb_purge(request.purge.va, request.purge.itir);
			request.type = IPC_PURGE_REPLY;
			ipc_send(node, &request);
			break;
		case IPC_PURGE_REPLY:
			ret = 1;
			break;
		case IPC_IPI:
			IPC_DEBUG("Rcv IPI %x\n", request.ipi.irq);
			deliver_interrupt(sc, request.ipi.irq);
			break;
		case IPC_STARTUP:
			smp_rendezvous_ip = request.startup.ip;
			smp_rendezvous_gp = request.startup.gp;
			ret = 1;
			break;
		case IPC_SSC:
			disk_req.addr = request.ssc.addr;
			disk_req.len = request.ssc.len;
			do_ssc(sc, request.ssc.fd, 0, (unsigned long)&disk_req, request.ssc.offset, request.subtype);
			request.type = IPC_SSC_REPLY;
			ipc_send(node, &request);
			break;
		case IPC_SSC_REPLY:
			ret = 1;
			break;
		case IPC_SHUTDOWN:
			ipc_send_reply(node, IPC_SSC_REPLY);
			exit(0);
			break;
		default:
			printf("Unexpected message %d\n", request.type);
		}
	}

	if (errno != EAGAIN)
		vm_perror("recv");

	return ret;
}

void
ipc_send(int node, struct request *request)
{
	if (node == cpu.id)
		vm_panic(NULL, "send to self\n");
	if (node >= ncpu)
		vm_panic(NULL, "Trying to access non-existant cpu, add more CPUs with -n <cpus>\n");

	IPC_DEBUG("send message %s -> %d\n", messages[request->type], node);
	if (send(sock[node], request, sizeof(*request), 0) != sizeof(*request))
		vm_perror("send");
	IPC_DEBUG("send message %s -> %d DONE\n", messages[request->type], node);
}

void
ipc_recv(struct sigcontext *sc)
{
	siginfo_t info;
	sigset_t sigset;
	unsigned long saved_psr;

	IPC_DEBUG("ipc_recv\n");

	/* rsm psr.i so that we don't inadvertantly deliver an interrupt */
	saved_psr = cpu.psr;
	cpu.psr &= ~PSR_I;

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGIPC);

	while (1)
	{
		if (sigwaitinfo(&sigset, &info) == -1)
			continue;
		if (handle_SIGIO_core(&info, sc))
			break;
	}

	cpu.psr = saved_psr;
	check_interrupts(sc);

	IPC_DEBUG("ipc_recv done\n");
}

void
ipc_call(struct sigcontext *sc, int node, struct request *request)
{
	/* since we're in a signal handler SIGALRM/SIGIO should be blocked already */
	IPC_DEBUG("ipc_call from %d to %d\n", cpu.id, node);
	ipc_send(node, request);
	IPC_DEBUG("ipc_call: recv\n");
	ipc_recv(sc);
}

void
ipc_broadcast(struct sigcontext *sc, unsigned long mask, struct request *request)
{
	int i;

	/* OPTIMISE ME */

	for (i = 0; i < cpus; i++)
	{
		if (mask & (1 << i))
		{
#if 0
			if (i == cpu.id)
				ipc_handle_async(-1, request);
			else
#endif
			if (i != cpu.id)
				ipc_call(sc, i, request);
		}
	}
}
