#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <elf.h>
#include <sys/mman.h>
#include <linux/stringify.h>
#include <sched.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <setjmp.h>
#include "debug.h"

#define NETWORK
#ifdef NETWORK
#include <sys/ioctl.h>
#include <net/if.h>
#include <linux/if_tun.h>
#endif

#ifndef PTRACE_VIRTUAL
#  define PTRACE_VIRTUAL 39
#endif

#include "vm.h"

#define roundup(x, y) (((x) + (y) - 1)&(~((y)-1)))

#ifndef MREMAP_FIXED
#  define MREMAP_FIXED 2
#endif

static inline void *vm_mremap(void *old_address, size_t old_size,
			size_t new_size, int flags, void *newaddr) {
	return (void *)syscall(__NR_mremap, old_address, old_size, new_size, flags, newaddr);
}

/* 
 * defaults
 */
int ncpu = 1;
unsigned long memsize = 512*MB;

struct termios orig_termios;
int restore_termios;

int keyboard_irq, network_irq;
unsigned char network_ethaddr[6];

char args[256];
int cpus;
unsigned long smp_rendezvous_ip, smp_rendezvous_gp;

extern struct cpu cpu;
int dumping;
jmp_buf continuation;

static unsigned long lowmem = ~0UL; /* where the kernel is loaded */
static unsigned long highmem = 0;

static void
set_affinity(int addr)
{
	cpu_set_t s;
	int i;
	int j;
	int k;

	sched_getaffinity(0, sizeof s, &s);
	/*
	 * How many available processors are there?
	 */
	for (j = i = 0; i < sizeof s; i++)
		if (CPU_ISSET(i, &s))
			++j;

	/*
	 * Step through to the addr th one and use it.
	 */
	for (k = i = 0; i < sizeof s; i++)
		if (CPU_ISSET(i, &s))
			if (k++ == addr % j)
				break;
	CPU_ZERO(&s);
	CPU_SET(i, &s);
	fprintf(stderr, "Allowing running on processor %d of %d\n",
		i, j);
	sched_setaffinity(0, sizeof s, &s);
}



pid_t vm_clone(void)
{
	return syscall(__NR_clone2, SIGCHLD|CLONE_FS|CLONE_FILES, NULL, NULL, NULL, NULL, NULL);
}

static void
cpu_init(int tapnum)
{
	int i, cpunumber = 0;

	cpus = NODES;

	for (i = 1; i < cpus; i++)
	{
		if (vm_clone() == 0)
		{
		  cpunumber = i;
			set_affinity(tapnum + i);
			break;
		}
	}
	/* 
	 * Mark data region as shared.
	 *
	 * Do this *after* the clone to get the right sharing.
	 * We want to share with multiplexed address spaces, not 
	 * with other processors.
	 */
	{
		extern char data[];
		unsigned long sd = (unsigned long)data & ~(pagesize - 1);
		long len = roundup((unsigned long)sbrk(0), pagesize)-sd;
		char *p;

		p = mmap((void *)NULL, len, 
			       PROT_READ|PROT_WRITE, 
			       MAP_SHARED|MAP_ANONYMOUS, 0, 0);
		if (p == MAP_FAILED){
			vm_perror("mmap failed\n");
			exit(1);
		}
		memcpy(p, (char *)sd, len);
		p = vm_mremap(p, len, len, MREMAP_MAYMOVE|MREMAP_FIXED, (char *)sd);
		if (p == MAP_FAILED)
			vm_perror("mremap");
	}

	mmap((void *)CPUADDR, pagesize*2, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_FIXED|MAP_ANONYMOUS, 0, 0);
	cpu.cfle = 1;
	cpu.psr = PSR_BN; /* Start in bank 1 */
	cpu.cr[CR_ITV] = 1UL << 16; /* Mask timer interrupts */
	cpu.id = cpunumber;
	cpu.cr[CR_LID] = cpunumber << 16;
}

static void
devices_init(unsigned tapnum)
{
	struct termios termios;
#ifdef NETWORK
	struct ifreq ifr;
	int sock;
#endif

	/* Init console */

	if (tcgetattr(STDIN_FILENO, &termios) != -1)
	{
		orig_termios = termios;
		termios.c_iflag &= ~(INLCR|ICRNL);
		termios.c_lflag &= ~(ICANON|ISIG|ECHO);
		if (tcsetattr(STDIN_FILENO, TCSANOW, &termios) != -1)
			restore_termios = 1;
	}

#ifdef NETWORK
	/* Init network */
	network_fd = open("/dev/net/tun", O_RDWR);
	if (network_fd != -1)
	{
		char name[IFNAMSIZ];
		snprintf(name, IFNAMSIZ, "tap%u", tapnum);
		memset(&ifr, 0, sizeof(ifr));
		ifr.ifr_flags = IFF_TAP|IFF_NO_PI;
		strncpy(ifr.ifr_name, name, IFNAMSIZ);
		ioctl(network_fd, TUNSETIFF, &ifr);
		ioctl(network_fd, TUNSETNOCSUM, 1);

		sock = socket(AF_INET, SOCK_DGRAM, 0);
		if (ioctl(sock, SIOCGIFHWADDR, &ifr) == -1)
			vm_perror("SIOCGIFHWADDR");
		memcpy(network_ethaddr, ifr.ifr_hwaddr.sa_data, sizeof(network_ethaddr));
		close(sock);

		IPC_DEBUG("network fd is %d\n", network_fd);
	} 
	else
	{
		fprintf(stderr, "No /dev/net/tun\n");
	}
#else
	network_fd = -1;
#endif
}

void
dump_regs(struct sigcontext *sc)
{
	unsigned long *prev_frame_base = (unsigned long *)((sc->sc_ar_bsp & ~0x1ff) - 512 - 512);
	unsigned long *alt_rbs_base = (unsigned long *)(sc->sc_rbs_base);
	int i;

	asm volatile(";; flushrs ;;");

#if 0
	sc->sc_gr[4] = read_register(sc, 4);
	sc->sc_gr[5] = read_register(sc, 5);
	sc->sc_gr[6] = read_register(sc, 6);
	sc->sc_gr[7] = read_register(sc, 7);
#endif
	fprintf(stderr, "ip   = %016lx\tpsr  = %016lx\tcfle = %d\n", sc->sc_ip, cpu.psr, cpu.cfle);
	fprintf(stderr, "cfm  = %016lx\tpfs  = %016lx\n", sc->sc_cfm, sc->sc_ar_pfs);
	fprintf(stderr, "bsp  = %016lx\trsc  = %016lx\n", sc->sc_ar_bsp, sc->sc_ar_rsc);
	fprintf(stderr, "arbs = %016lx\tldrs = %016lx\n", sc->sc_rbs_base, sc->sc_loadrs);
	fprintf(stderr, "pr   = %016lx\tnat  = %016lx\n", sc->sc_pr, sc->sc_nat);
	fprintf(stderr, "rnat = %016lx\tunat = %016lx\n", sc->sc_ar_rnat, sc->sc_ar_unat);
	for (i = 0; i < 32; i += 2)
		fprintf(stderr, "r%-2d  = %016lx\tr%-2d  = %016lx\n", i, sc->sc_gr[i], i+1, sc->sc_gr[i+1]);
	for (i = 0; i < 8; i += 2)
		fprintf(stderr, "b%-2d  = %016lx\tb%-2d  = %016lx\n", i, sc->sc_br[i], i+1, sc->sc_br[i+1]);
	for (i = 0; i < 64; i += 2)
		fprintf(stderr, "f%-2d  = %016lx%016lx\tf%-2d  = %016lx%016lx\n", i, sc->sc_fr[i].u.bits[1], sc->sc_fr[i].u.bits[0], i+1, sc->sc_fr[i+1].u.bits[1], sc->sc_fr[i+1].u.bits[0]);

	fprintf(stderr, "dcr  = %016lx\titm  = %016lx\n", cpu.cr[CR_DCR], cpu.cr[CR_ITM]);
	fprintf(stderr, "iva  = %016lx\tpta  = %016lx\n", cpu.cr[CR_IVA], cpu.cr[CR_PTA]);
	fprintf(stderr, "ipsr = %016lx\tisr  = %016lx\n", cpu.cr[CR_IPSR], cpu.cr[CR_ISR]);
	fprintf(stderr, "iip  = %016lx\tifa  = %016lx\n", cpu.cr[CR_IIP], cpu.cr[CR_IFA]);
	fprintf(stderr, "itir = %016lx\tiipa = %016lx\n", cpu.cr[CR_ITIR], cpu.cr[CR_IIPA]);
	fprintf(stderr, "ifs  = %016lx\tiim  = %016lx\n", cpu.cr[CR_IFS], cpu.cr[CR_IIM]);
	fprintf(stderr, "iha  = %016lx\n", cpu.cr[CR_IHA]);

	dumping = 1;
	if (setjmp(continuation) == 0) {
		fprintf(stderr, "RSE backing store:\n");
		for (i = 0; i < 256; i += 2)
			fprintf(stderr, "%016lx: %016lx %016lx\n", (unsigned long)&prev_frame_base[i], prev_frame_base[i], prev_frame_base[i+1]);
	}
	if (setjmp(continuation) == 0) {
		fprintf(stderr, "Alternate backing store:\n");
		for (i = 0; i < (sc->sc_loadrs >> 19); i += 2)
			fprintf(stderr, "%016lx: %016lx %016lx\n", (unsigned long)&alt_rbs_base[i], alt_rbs_base[i], alt_rbs_base[i+1]);
	}
	dumping = 0;
	for (i = 0; i < 4; i++)
		fprintf(stderr, "rr[%i]=%016lx\trr[%i]=%016lx\n",
			i, cpu.rr[i], i+4, cpu.rr[i+4]);
	int bank = cpu.psr & PSR_BN ? 16 : 0;
	fprintf(stderr, "Register Bank %x\n", bank / 16);
	for (i = 0; i < 16; i += 2)
	  fprintf(stderr, "r%-2d  = %016lx\tr%-2d  = %016lx\n", 
		  i + 16, cpu.saved_bank[i + bank],
		  i + 17, cpu.saved_bank[i + bank + 1]);
}

void
dump_cpu(void)
{
#define CR(x) fprintf(stderr, "CR[%d] = %lx\t", x, cpu.cr[x]);
//#ifndef AFTERBURN

	CR(100); CR(101); CR(102);
	fprintf(stderr, "\n");
	CR(103); CR(104); CR(105);
	fprintf(stderr, "\n");
	CR(106); CR(107);
	fprintf(stderr, "\n");
//#endif
	CR(68); CR(69); CR(70); CR(71);
	fprintf(stderr, "VIRTUAL PSR: 0x%lx\nSAVED GR:\n\t r4=0x%lx\tr6=0x%lx\n\t r5=0x%lx\tr7=0x%lx\n\t b2=0x%lx\tb6=0x%lx\n",
		cpu.psr, 
		cpu.gr[0], cpu.gr[2], cpu.gr[1], cpu.gr[3], cpu.saved_b2, cpu.saved_b6);

}

#if 1
/* produce a hex dump */
void
hexdump(unsigned char *p, int len)
{
        unsigned char *line = p;
        int i, thisline, offset = 0;

        while (offset < len)
        {
                fprintf(stderr, "%04x ", offset);
                thisline = len - offset;
                if (thisline > 16)
                        thisline = 16;

                for (i = 0; i < thisline; i++)
                        fprintf(stderr, "%02x ", line[i]);

                for (; i < 16; i++)
                        fprintf(stderr, "   ");

		for (i = 0; i < thisline; i++)
			fprintf(stderr, "%c", (line[i] >= 0x20 && line[i] < 0x7f) ? line[i] : '.');

		fprintf(stderr, "\n");
                offset += thisline;
                line += thisline;
        }
}

unsigned long breakpoint_address;
unsigned long breakpoint_contents[2];

void
debug(struct sigcontext *sc)
{
	char line[256];
	unsigned long address;

	if (sc && sc->sc_ip == breakpoint_address)
	{
		fprintf(stderr, "breakpoint %lx hit\n", breakpoint_address);
		address = breakpoint_address;
		*(unsigned long *)address = breakpoint_contents[0];
		*(unsigned long *)(address + 8) = breakpoint_contents[1];
	}

	while (1)
	{
		fprintf(stderr, "> ");
		if (fgets(line, sizeof(line), stdin) == NULL)
			return;
		if (!line[0])
			continue;
		switch (line[0])
		{
			case 'b':
				if (line[1] != ' ')
					continue;

				address = strtoul(&line[2], NULL, 16);
				fprintf(stderr, "breakpoint %lx\n", address);
				breakpoint_address = address;
				breakpoint_contents[0] = *(unsigned long *)address;
				breakpoint_contents[1] = *(unsigned long *)(address + 8);
				*(unsigned long *)address = 0;
				*(unsigned long *)(address + 8) = 0;
				asm volatile("fc.i %0" : : "r"(address));
				break;

			case 'c':
				fprintf(stderr, "continue\n");
				return;

			case 'd':
				if (line[1] != ' ')
					continue;

				address = strtoul(&line[2], NULL, 16);
				fprintf(stderr, "dump %lx\n", address);
				hexdump((unsigned char *)address, 16*16);
				break;
		}
	}
}
#endif

void
dump_maps(void)
{
	char buffer[4096];
	int fd, len;

	fflush(stdout);
	fd = open("/proc/self/maps", O_RDONLY);
	while ((len = read(fd, buffer, sizeof(buffer))) > 0)
		fwrite(buffer, 1, len, stderr);
	close(fd);
}

void 
dump_stats(void)
{
#ifdef EMULATOR_STAT

#define ARG(x) __stringify(x), stats.x
#define FMT    "%12s: %06lu\t"
	fprintf(stderr, "Emulator statistics\n"
		FMT FMT FMT FMT "\n" FMT FMT FMT FMT "\n"
		FMT FMT FMT FMT "\n" FMT FMT FMT FMT "\n"
		FMT FMT FMT FMT "\n" FMT FMT FMT FMT "\n"
		FMT FMT FMT FMT "\n" FMT FMT FMT FMT "\n"
		FMT FMT FMT FMT "\n" 
		FMT FMT FMT FMT "\n" FMT FMT FMT FMT "\n" FMT "\n",

		ARG(nbreak), ARG(ssm), ARG(rsm), ARG(mov_to_rr),
		ARG(mov_from_rr), ARG(mov_to_pkr), ARG(ptc_l),
		ARG(ptc_g), ARG(ptc_e), ARG(itr), ARG(nthash), ARG(nttag),
		ARG(mov_from_pmd), ARG(mov_from_pmc),
		ARG(ntpa), ARG(mov_from_cr), ARG(nread_psr),
		ARG(mov_to_ar), ARG(mov_to_psr), ARG(itc_d),
		ARG(itc_i), ARG(shrinkframes), ARG(rfi), ARG(bsw), ARG(redirects),
		ARG(faults), ARG(sigill), ARG(breaks), ARG(ticks),
		ARG(sigio), ARG(fpe),
		ARG(ssc), ARG(break_cover), ARG(break_dfh), ARG(break_halt),
		ARG(break_syscall), ARG(break_tpa), ARG(break_tlbop),
		ARG(break_rr), ARG(break_itm), ARG(break_rfi),
		ARG(break_irq), ARG(break_ivr), ARG(break_debug), ARG(break_pause)
);
#endif
}

void
vm_ssc_shutdown(void)
{
	struct request req;
	req.type = IPC_SHUTDOWN;
	ipc_broadcast(NULL, (unsigned long)-1, &req);
	vm_shutdown();
}

void vm_shutdown(void)
{	
	if (restore_termios)
		tcsetattr(STDIN_FILENO, TCSANOW, &orig_termios);

	aio_shutdown();
	dump_stats();
#if DEBUG & DEBUG_CTXT_BIT
	dump_ctx();
#endif
	exit(0);
}

void
vm_panic(struct sigcontext *sc, const char *fmt, ...)
{
	va_list va;
	
	/* fprintf(stderr, "r39 = %lx\n", read_register(sc, 39)); */
	fprintf(stderr, "PANIC in VMM: ");
	msync(0UL, ~0UL, MS_SYNC);
	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);

	if (sc != NULL) {
//		backtrace(sc);
		dump_regs(sc);
	}
	dump_cpu();
	dump_maps();
	vm_shutdown();
	exit(0);
}

/*
 * perror and strerror link dynamic libraries at run time with I18N localised code
 * so we can't use them.
 */
void
vm_perror(char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);

	fprintf(stderr, ": %s\n", sys_errlist[errno]);
}

static void *
load_kernel(char *file)
{
	Elf64_Ehdr hdr;
	Elf64_Phdr *phdrs, *phdr;
	int fd, i;

	fd = open(file, O_RDONLY);
	if (fd == -1) {
		vm_perror(file);
		return NULL;
	}

	read(fd, &hdr, sizeof(Elf64_Ehdr));

	if (hdr.e_ident[0] != 0x7f || strncmp((char *)hdr.e_ident + 1, "ELF", 3) != 0) {
		fprintf(stderr, "%s is not an ELF file\n", file);
		exit(1);
	}
	if (hdr.e_type != ET_EXEC) {
		fprintf(stderr, "%s is not an ELF executable\n", file);
		exit(1);
	}
	if (hdr.e_machine != EM_IA_64) {
		fprintf(stderr, "kernel not for this processor\n");
		exit(1);
	}

	lseek(fd, hdr.e_ehsize, SEEK_SET);
	phdr = phdrs = malloc(hdr.e_phnum * hdr.e_phentsize);
	read(fd, phdrs, hdr.e_phnum * hdr.e_phentsize);

	for (i = 0; i < hdr.e_phnum; i++)
	{
		if (phdr->p_paddr < lowmem)
			lowmem = phdr->p_paddr;
		if (phdr->p_paddr + phdr->p_memsz > highmem)
			highmem = phdr->p_paddr + phdr->p_memsz;
		lseek(fd, phdr->p_offset, SEEK_SET);
		read(fd, (void *)(phdr->p_paddr), phdr->p_filesz);
		memset((char *)phdr->p_paddr + phdr->p_filesz, 0,
		       phdr->p_memsz - phdr->p_filesz);
		phdr = (Elf64_Phdr *)((char *)phdr + hdr.e_phentsize);
	}
	free(phdrs);
	close(fd);

	return (void *)hdr.e_entry;
}


void
Usage(void)
{
	fprintf(stderr, "Usage: vmm [ -t tapnumber] [-m memsize[kMG]] [-n ncpu] [-k] vmlinux args ...\n");
	exit(1);
}


int
main(int argc, char *argv[])
{
	static char outbuf[BUFSIZ];
	void *entry;
	int i, offset = 0;
	int opt;
	unsigned tapnum = 0;
	mainpid = getpid();
	int keep_files = 0;

	cpu_freq = get_itc_freq();

	while ((opt = getopt(argc, argv, "t:n:m:k")) != -1)
	{
		unsigned long x;
		char *p;

		switch (opt) {
		case 't':	/* Network device to use */
			x = strtoul(optarg, &p, 10);
			if (*p || x > 16)
				Usage();
			tapnum = x;
			break;
		case 'n':	/* ncpu */
			x = strtoul(optarg, &p, 10);
			if (*p || x > MAXCPU)
				Usage();
			ncpu = x;
			break;
		case 'm':	/* Memory */
			x = strtoul(optarg, &p, 10);
			switch (*p) {
			case 'k':
			case 'K':
				x *= KB;
				break;
			case 'm':
			case 'M':
				x *= MB;
				break;
			case 'g':
			case 'G':
				x *= GB;
				break;
			default:
				Usage();
			}
			memsize = x;
			break;
		case 'k':	/* Keep memory files */
			keep_files = 1;
			break;

		default:
			Usage();
		}
	}

	/* avoid stdio allocating its own buffer */
	setvbuf(stdout, outbuf, _IOLBF, sizeof(outbuf));

	argc -= optind - 1;
	argv += optind - 1;

	for (i = 2; i < argc; i++)
	{
		strncpy(&args[offset], argv[i], sizeof(args)-offset);
		offset += strlen(argv[i]);
		args[offset++] = (i < argc-1) ? ' ' : 0;
	}

	fprintf(stderr, "args: %s\n", args);

	mm_init(keep_files);
	cpu_init(tapnum);
	/*
	 * Now running all processors in parallel
	 */
	aio_init(keep_files);
	syscall_trace();
	as_init();
	signal_init();
	ipc_init();

	fprintf(stderr, "%d: CPU %d\n", getpid(), cpu.id);
	if (cpu.id == 0)
	{
		void *bp;

		set_affinity(tapnum);
		devices_init(tapnum);
		entry = load_kernel(argv[1]);
		fprintf(stderr, "entry = %p\n", entry);
		if (entry == NULL)
		{
			vm_shutdown();
			return 1;
		}

		bp = get_bootparams(lowmem, highmem, args);
		fprintf(stderr, "Booting the kernel\n");
		asm volatile("mov r28=%1; br.sptk.few %0" :: "b"(entry), "r"(bp));
		vm_panic(NULL, "kernel returned\n");
	}

	goto *(void *)smp_rendezvous_ip;
}

