#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <asm/unistd.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include "vm.h"
#include "debug.h"

#ifndef  __NR_switch_vmm
#error "Include files do not include multi-as system call: check KERNEL Makefile variable"
#endif


static int last_asid = 0;

#ifdef OPTIMISE_USE_MULTI_AS
/*
 * Create a new address space
 */
static inline long
new_as(void)
{
	long n = syscall(__NR_new_vmm);
	return n;
}

/*
 * Switch to a new (previously created) address space
 */
static inline long
switch_as(int as)
{
	long n;
	static int current_asid;

	if (current_asid == as)
		return 0;

	n =  syscall(__NR_switch_vmm, as);
	if (n < 0)
		vm_perror("switch_as");
	else
		current_asid = as;
	return n;
}

#else
static inline long new_as(void)
{
	return -1;
}

static inline long
switch_as(int as)
{
	return -1;
}
#endif

struct managed_page {
	unsigned long copyset;
	unsigned int owner;
	unsigned int invalidate_in_progress;
};

struct page {
	struct pte *mappings;
	int prot;
};

struct pte {
	unsigned long va;
	unsigned long pa;
	unsigned int prot;
	struct pte *hash_next;
	struct pte *pa_next;
};

struct queued_page {
	unsigned long pa;
	int is_write;
	int node;
};

/*
 * These bits are carefully chosen to 
 * avoid the PROT_READ/PROT_WRITE/PROT_EXEC
 * bits defined in /usr/include/asm-generic/mman.h
 */
#define PROT_ACCESSED	0x08
#define PROT_DIRTY	0x10
#define PROT_PINNED	0x20
#ifndef PROT_SEM
# define PROT_SEM 0
#endif

#define PROT_STDMASK (PROT_READ|PROT_WRITE|PROT_EXEC|PROT_SEM)

#if (PROT_ACCESSED|PROT_DIRTY|PROT_PINNED) & (PROT_READ|PROT_WRITE|PROT_EXEC|PROT_SEM|PROT_GROWSDOWN|PROT_GROWSUP)
# error "PROT_XXX bits need redefining
#endif

#define NCONTEXTS 16
#define ASID_HIST 15
static unsigned ncontexts;
static unsigned asid_hist;

/*
 * Recently used ASID relies on the history being small
 * otherwise it'll perform very poorly.
 * in any case it *must* be less than NCONTEXTS
 */
#if ASID_HIST >= NCONTEXTS
# error "ASID_HIST must be less than NCONTEXTS"
#endif

/* 64 pretty arbitrary definition of `small' */
/* Kernel support may only give 8 or 16 */
#if NCONTEXTS > 64
# error "NCONTEXTS must be <= 64"
#endif

#ifdef DSM
/* physical pages managed by this node - owner/copyset information */
struct managed_page managed_pages[(MEM_SIZE/NODES)>>PAGE_SHIFT];

/* allows locating virtual mappings for a physical page */
struct page *pages;
#endif

#if 0
#define PGTABLE_ORDER   (20)
#define PGTABLE_ENTRIES (1 << PGTABLE_ORDER)
#define PGTABLE_MASK    (PGTABLE_ENTRIES-1)
#define HASH(va)  ((va >> 61 << (PGTABLE_ORDER-3)) | ((va >> PAGE_SHIFT) & (PGTABLE_MASK>>3)))
#endif

unsigned long pagesize;
#define RGN(x) ((x)>>61)

/* 
 * Linux steals regions 5, 6 and 7;
 * so we have only 5 regions for the guest to use 
 */
#define HI_REGION 5 

#define PGTABLE_ENTRIES		1024
#define PGTABLE_MASK		(PGTABLE_ENTRIES-1)
static inline unsigned long HASH(unsigned long va)
{
	return ((va >> PAGE_SHIFT) & PGTABLE_MASK);
}

struct pte *pagetables[NCONTEXTS][HI_REGION][PGTABLE_ENTRIES];
struct pte **pagetable[HI_REGION] = { pagetables[0][0], 
				      pagetables[0][1], 
				      pagetables[0][2],
				      pagetables[0][3], 
				      pagetables[0][4]};
/*
 * Keep track of address space ID for each region and each context.
 * asid contains the value of the address space ID (i.e., rr >> 8)
 * Query: maybe a hash table instead of linear search?
 * Query: Unify all regions for this (we can't switch individual RRs yet)
 */
int asid[HI_REGION][NCONTEXTS];

#if ASID_HIST
/*
 * context_hist keeps track of the last ASID_HIST contexts used
 * for each region.
 */
int context_hist[HI_REGION][ASID_HIST]; /* recently used contexts */
int context_hist_idx[HI_REGION]; /* index into context_hist -- the next one to displace */
#endif
int last_context[HI_REGION];  /* Last context number displaced */

int mem_fd;	/* File descriptor of file that is physical memory */

unsigned long last_invalidate_address;

#define PTE_POOL_SIZE (1024*1024)
static struct pte pte_pool[PTE_POOL_SIZE];
static struct pte *next_pte = pte_pool;
#define CHECK_PTE(pte) do { if ((pte) &&((pte) < pte_pool || (pte) >= &pte_pool[PTE_POOL_SIZE])) vm_panic(NULL, "pte out of range"); } while (0)

static struct pte *
pte_alloc(void)
{
	struct pte *pte = next_pte;
	next_pte = pte->pa_next;
	return pte;
}

static void
pte_free(struct pte *pte)
{
	pte->pa_next = next_pte;
	next_pte = pte;
}

/*
 * Remove mappings from the page table, and 
 * the corresponding mmaps onto the physical memory
 */
static void
mappings_remove(unsigned long va, long size, int force)
{
	struct pte *pte, **ppte;

	while (size > 0)
	{
		ppte = &pagetable[RGN(va)][HASH(va)];
		for (; (pte = *ppte) != NULL; ppte = &(pte->hash_next))
		{	
			CHECK_PTE(pte);
			if (pte->va == va)
			{
				if (!(pte->prot & PROT_PINNED) || force)
				{
					/* remove from hash list */
					*ppte = pte->hash_next;
					CHECK_PTE(*ppte);

#if DSM
					/* remove from pa list */
					ppte = &pages[pte->pa >> PAGE_SHIFT].mappings;
					for (; *ppte != NULL; ppte = &((*ppte)->pa_next))
					{
						if (*ppte == pte)
						{
							*ppte = pte->pa_next;
							break;
						}
					}
#endif
					pte_free(pte);
				}
				break;
			}
		}

		va += pagesize;
		size -= pagesize;
	}
	munmap((void *)va, size);
}

static void
mappings_add(unsigned long va, unsigned long pa, long size, int prot)
{
	struct pte *pte;

	while (size > 0)
	{
		pte = pte_alloc();
		if (pte == NULL)
			fprintf(stderr, "BAD! OUT OF PTES!\n");
		CHECK_PTE(pte);
		pte->va = va;
		pte->pa = pa;
		pte->prot = prot;
		pte->hash_next = pagetable[RGN(va)][HASH(va)];
		CHECK_PTE(pte->hash_next);
		pagetable[RGN(va)][HASH(va)] = pte;
#if DSM
		pte->pa_next = pages[pa >> PAGE_SHIFT].mappings;
		pages[pa >> PAGE_SHIFT].mappings = pte;
#endif

		va += pagesize;
		pa += pagesize;
		size -= pagesize;
	}
}

int
tlb_insert(unsigned long pte, int pinned)
{
	unsigned long size = 1UL << BITS(cpu.cr[CR_ITIR], 2, 6);
	unsigned long va = cpu.cr[CR_IFA] & ~(size - 1);
	unsigned long pa = (BITS(pte, 12, 38) << 12) & ~(size - 1);
	unsigned long ar = BITS(pte, 9, 3);
 	int writeable, executable, accessed, dirty;
	int prot;
	void *ret;

	DEBUG_MM("Mapping from %lx to %lx, size %lx (ITIR = %lx)\n", va, pa, size, cpu.cr[CR_ITIR]);

	if ((RGN(va) == 0 && va >= MEM_START) || ((va < MEM_START) && (va + size >= MEM_START)))
	{
		if (pa != va)
			DEBUG_MM("Mapping into region 0 not allowed\n");
		return pa != va;
	}

	if (!(pte & 1)) /* present */
		return 0;

	if (ar == 4)
		printf("WARNING: ar == 4\n");

	writeable = (ar == 2) || (ar == 3) || (ar == 6);
	executable = (ar == 1) || (ar == 3) || (ar >= 5);
	accessed = BITS(pte, 5, 1);
	dirty = BITS(pte, 6, 1);

	prot = PROT_READ;
	if (writeable)
		prot |= PROT_WRITE;
	if (executable)
		prot |= PROT_EXEC;
	if (accessed)
		prot |= PROT_ACCESSED;
	if (dirty)
		prot |= PROT_DIRTY;
	if (pinned)
		prot |= PROT_PINNED;
	mappings_remove(va, size, 1);
	mappings_add(va, pa, size, prot);

#if DSM
	/* make sure actual protection is a subset of that on the physical pages */
	prot &= pages[pa >> PAGE_SHIFT].prot;
#endif
	prot &= PROT_STDMASK;
	if (!accessed)
		prot &= ~(PROT_READ|PROT_EXEC|PROT_WRITE);
	else if (!dirty)
		prot &= ~PROT_WRITE;

	if (pinned)
		ret = mmapall((void *)va, size, prot, MAP_FIXED|MAP_SHARED, mem_fd, pa);
	else
		ret = mmap((void *)va, size, prot, MAP_FIXED|MAP_SHARED, mem_fd, pa);
	if (ret == MAP_FAILED) {
		fprintf(stderr, "Cannot map from %p,%p to %p--- ",
			(char *)va, ((char *)va)+size-1, (char *)pa);
		vm_perror("mmap");
		return 1;
	}

	return 0;
}

void
tlb_purge(unsigned long va, unsigned long itir)
{
	unsigned long size = 1UL << BITS(itir, 2, 6);
	va &= ~(size - 1);

	DEBUG_MM("Unmapping from %lx, size %lx\n", va, size);

	if (RGN(va) == 0)
	{
		DEBUG_MM("Purging region 0 not allowed\n");
		return;
	}

	mappings_remove(va, size, 0);
	/* Make sure it's done! */
	munmap((void *)va, size);
}

void
tlb_purge_global(struct sigcontext *sc, unsigned long va, unsigned long itir)
{
#if 0
	struct request request;
#endif

	tlb_purge(va, itir);
#if 0
	request.type = IPC_PURGE;
	request.purge.va = va;
	request.purge.itir = itir;
	ipc_broadcast(sc, (unsigned long)-1, &request);
#endif
}

void
tlb_purge_region(unsigned int region)
{
	struct pte *pte;
	int i;

	DEBUG_MM("Purging region %d\n", region);

	if ((region == 0) || (region >= 4))
	{
		printf("Not purging region %d\n", region);
		return;
	}

	for (i = 0; i < PGTABLE_ENTRIES; i++)
	{
		for (pte = pagetable[region][i]; pte != NULL; pte = pte->hash_next)
		{
			CHECK_PTE(pte);
			if (RGN(pte->va) != region)
				vm_panic(NULL, "region %d: va %lx doesn't match region!\n", region, pte->va);
			if (!(pte->prot & PROT_PINNED))
				mappings_remove(pte->va, pagesize, 0);
		}
	}

	/* Make sure it's done! */
	munmap((void *)((unsigned long)region << 61), 1UL << 61);
}

void
tlb_purge_all(void)
{
	struct pte *pte;
	int i;
	int j;

	DEBUG_MM("Purging TLB\n");

	for (j = 0; j < HI_REGION-1; j++)
		for (i = 0; i < PGTABLE_ENTRIES; i++)
		{
			for (pte = pagetable[j][i]; pte != NULL; pte = pte->hash_next)
			{
				if (!(pte->prot & PROT_PINNED))
					mappings_remove(pte->va, pagesize, 0);
			}
		}

	/* Make sure even pesky mappings like our ex-stack are gone */
	munmap((void *)(1UL << 61), 3UL << 61);
}

unsigned long
thash(unsigned long addr)
{
	unsigned long vhpt_mask, vhpt_base, vhpt_offset, region;

	region = RGN(addr);
	vhpt_mask = (1UL << BITS(cpu.cr[CR_PTA], 2, 6)) - 1;
	/* vhpt_base = (cpu.cr[CR_PTA] & 0x1fffffff80000000UL) & ~vhpt_mask; 0x1000000000UL */
	vhpt_base = BITS(cpu.cr[CR_PTA], 15, 49);
	vhpt_offset = ((addr << 3) >> 3) >> BITS(cpu.rr[region], 2, 6);
	if (BITS(cpu.cr[CR_PTA], 8, 1)) /* long format */
		return (vhpt_base | ((vhpt_offset << 5) & vhpt_mask));
	return ((region << 61) | ((vhpt_base << 3) >> 3) | ((vhpt_offset << 3) & vhpt_mask));
}

unsigned long
ttag(unsigned long addr)
{
	unsigned long vhpt_offset, region;

	region = RGN(addr);
	vhpt_offset = ((addr << 3) >> 3) >> BITS(cpu.rr[region], 2, 6);
	return (vhpt_offset ^ BITS(cpu.rr[region], 8, 18));
}

unsigned long
tpa(unsigned long addr)
{
	unsigned long page;
	struct pte *pte;
	extern void _text, _end;
	/* much of region 0 is identity mapped */ 
	if (RGN(addr) == 0 && (
		    (addr >= MEM_START && addr < MEM_START+MEM_SIZE) 
		    ||
		    (addr >= (unsigned long)&_text && addr < (unsigned long)&_end))) {
		DEBUG_MM("Identity mapped TPA %lx->%lx\n", addr, addr);
		return addr;
	}

	page = addr & ~PAGE_MASK;
	pte = pagetable[RGN(addr)][HASH(addr)];
	for (; pte != NULL; pte = pte->hash_next)
	{
		if (pte->va == page) {
			DEBUG_MM("tpa: va %lx -> pa %lx\n",
				addr, pte->pa + (addr & PAGE_MASK));
			return pte->pa + (addr & PAGE_MASK);
		}
	}
	DEBUG_MM("TPA %lx returns 0\n", addr);
	return 0;
}

static inline unsigned long round_up(unsigned long x, unsigned y)
{
	return (x + (y-1)) & ~(y-1);
}


void
remap_region(unsigned region)
{
	int i;
	struct pte *pte;
	struct pte **basepte;

	CTXT_DEBUG("Remap_region %d\n", region);

	basepte = pagetable[region];
	if (region) {
		CTXT_DEBUG("munmap region %d\n", region);
		munmap((void *)((unsigned long)region << 61), 1UL<<61);
	}
	for (i = 0; i < PGTABLE_ENTRIES; i++) {
 		for (pte = basepte[i]; pte; pte = pte->hash_next) {
			mmap((void *)pte->va, pagesize, pte->prot, MAP_FIXED|MAP_SHARED, mem_fd, pte->pa);
		}
	}
}

/*
 * Set the pagetable up, and adjust the address space ID history
 */
void 
set_pgtable(int index, int new_asid, unsigned region)
{
	int i;

	/*
	 * TODO: Pinned mappings aren't accounted for here.
	 */
	pagetable[region] = pagetables[index][region];
	asid[region][index] = new_asid;

#if ASID_HIST
	if (region == 1){
		assert(index < ncontexts);
		CTXT_DEBUG("Switching to kernel context %d\n", index);
		DUMP_MAPS();
		switch_as(index);
		last_asid = index;
		DUMP_CTX();
		DUMP_MAPS();
	}

	/*
	 * Is the asid already there?
	 */
	for (i = 0; i < asid_hist; i++)
		if (context_hist[region][i] == index) {
			CTXT_DEBUG("set_pgtable: Found index %d in history at index %d\n",
				   index, i);
			return;
		}
	context_hist_idx[region] += 1;
	context_hist_idx[region] %= asid_hist;
	CTXT_DEBUG("Setting context_hist[%d, %d] to %d\n", region, context_hist_idx[region], index);
	context_hist[region][context_hist_idx[region]] = index;
#endif
}

void
set_asid(int next_asid, unsigned region)
{
	int i;
	unsigned long used;

	CTXT_DEBUG("Setting region %d to asid %x\n", region, next_asid);
	for (i = 0; i < ncontexts; i++) {
		if (asid[region][i] == next_asid) {
			CTXT_DEBUG("Found ASID in context %d\n", i);
			set_pgtable(i, next_asid, region);
#ifndef OPTIMISE_USE_MULTI_AS
			remap_region(region);
#endif
			return;
		}
	}

	/*
	 * We don't have a cached copy; find a not-recently-used 
	 * context, initialise it and set it up.
	 * The loop is guaranteed to end iff ASID_HIST < ncontexts
	 */
	used = 0;
#if ASID_HIST
	for (i = 0; i < asid_hist; i++)
		used |= (1 << context_hist[region][i]);
#endif
	for (i = 0; i < ncontexts; i++)
		if (used & (1 << last_context[region])) {
			last_context[region]++;
			if (last_context[region] == ncontexts)
				last_context[region] = 0;
		} else
			break;

	CTXT_DEBUG("Using context %d; purging old contents\n", last_context[region]);
	set_pgtable(last_context[region], next_asid, region);
	tlb_purge_region(region);
}

char VM_filename[32];

void
mm_init(int keep_files)
{

	int prot;
	char *p;
	int i;

	char *filename = VM_filename;

	sprintf(filename, "/tmp/vm_mem.%d", getpid());
	pagesize = getpagesize();
	printf("pagesize = %ld\n", pagesize);
	if (pagesize != (1<<PAGE_SHIFT)){
		fprintf(stderr, "Page size doesn't match guest pagesize: bailing out\n");
		exit(1);
	}
#ifdef DSM
	pages = mmap((void *)round_up(MEM_START+MEM_SIZE, pagesize), 
		     round_up((MEM_SIZE>>PAGE_SHIFT) * sizeof (*pages), pagesize), 
		     PROT_READ|PROT_WRITE,
		     MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, 0, 0);

	if (pages == MAP_FAILED) {
		vm_perror("mmap pages");
		exit (1);
	}
#endif
	mem_fd = open(filename, O_RDWR|O_CREAT|O_TRUNC, 0600);
	ftruncate(mem_fd, MEM_SIZE);

//	prot = (cpu.id == 0) ? (PROT_READ|PROT_WRITE|PROT_EXEC) : PROT_NONE;
	prot = PROT_READ|PROT_WRITE|PROT_EXEC;
#if DSM
	for (i = 0; i < (MEM_SIZE>>PAGE_SHIFT); i++)
		pages[i].prot = prot;
#endif
	if ((p = mmap((void *)MEM_START, MEM_SIZE-MEM_START, prot, MAP_FIXED|MAP_SHARED, mem_fd, MEM_START)) == (void *)-1) {
		vm_perror("mmap");
		unlink(filename);
		exit(1);
	}
	fprintf(stderr, "Physical memory at %p -> %p\n", p, p + MEM_SIZE - MEM_START);
	if (!keep_files)
		unlink(filename);

	for (i = 0; i < PTE_POOL_SIZE-1; i++)
		pte_pool[i].pa_next = &pte_pool[i+1];
}

#ifdef DSM
#define MAX_QUEUE 16
struct queued_page request_queue[MAX_QUEUE];
unsigned int rq_lo;
unsigned int rq_hi;

void
enqueue_request(int node, unsigned long pa, int is_write)
{
	unsigned int next_hi = (rq_hi + 1) % MAX_QUEUE;
	if (next_hi == rq_lo)
	{
		DEBUG_DSM("[%d] DROPPING REQUEST\n", getpid());
		return;
	}

	DEBUG_DSM("[%d] ENQUEUING REQUEST %d\n", getpid(), rq_hi);
	request_queue[rq_hi].node = node;
	request_queue[rq_hi].pa = pa;
	request_queue[rq_hi].is_write = is_write;
	rq_hi = next_hi;
}

void
process_request_queue(struct sigcontext *sc)
{
	unsigned int orig_hi = rq_hi;

	DEBUG_DSM("[%d] PROCESSING REQUEST QUEUE\n", getpid());
	while (rq_lo != orig_hi)
	{
		DEBUG_DSM("[%d] PROCESSING REQUEST %d\n", getpid(), rq_lo);
		mm_handle_fetch(sc, request_queue[rq_lo].node,
				request_queue[rq_lo].pa,
				request_queue[rq_lo].is_write);
		rq_lo = (rq_lo + 1) % MAX_QUEUE;
	}
}


void
mm_chmod(unsigned long pa, int prot)
{
	struct pte *pte;

	DEBUG_DSM("[%d] mm_chmod: %lx%s%s\n", getpid(), pa,
			(prot & PROT_READ) ? " read" : "", (prot & PROT_WRITE) ? " write" : "");

	for (pte = pages[pa >> PAGE_SHIFT].mappings; pte != NULL; pte = pte->pa_next)
		mprotect((void *)(pte->va), pagesize, pte->prot & prot);

	pages[pa >> PAGE_SHIFT].prot = prot;
	mprotect((void *)pa, pagesize, prot);
}

void
mm_handle_invalidate(int node, unsigned long pa, int is_write)
{
	struct request request;

	DEBUG_DSM("[%d] mm_handle_invalidate: %lx %s\n", getpid(), pa, is_write ? "write" : "read");

	mm_chmod(pa, is_write ? PROT_NONE : (PROT_READ|PROT_EXEC));

	last_invalidate_address = pa;
	request.type = IPC_INVALIDATE_REPLY;
	request.invalidate.pa = pa;
	request.invalidate.is_write = is_write;
	ipc_send(node, &request);
}

void
mm_handle_fetch(struct sigcontext *sc, int node, unsigned long pa, int is_write)
{
	struct request request;
	struct managed_page *page;
	int owner, manager;

	DEBUG_DSM("[%d] mm_handle_fetch: %lx %s\n", getpid(), pa, is_write ? "write" : "read");

	owner = manager = pa >> NODE_SHIFT;
	if (manager == cpu.id)
	{
		page = &managed_pages[(pa & NODE_MASK) >> PAGE_SHIFT];
		owner = page->owner;

		if (page->invalidate_in_progress)
		{
			enqueue_request(node, pa, is_write);
			return;
		}

		if (owner != cpu.id)
		{
			/* get ownership */
			DEBUG_DSM("[%d] need ownership\n", getpid());
			mm_fetch(sc, pa, is_write, pa);
		}

		if (is_write)
		{
			DEBUG_DSM("[%d] invalidating all other copies\n", getpid());

			/* invalidate all other copies, give ownership to requestor */
			assert(page->invalidate_in_progress == 0);
			page->invalidate_in_progress = 1;
			request.type = IPC_INVALIDATE;
			request.invalidate.pa = pa;
			request.invalidate.is_write = is_write;
			ipc_broadcast(sc, page->copyset & ~((1 << cpu.id) | (1 << node)), &request);
			assert(page->invalidate_in_progress == 1);
			page->invalidate_in_progress = 0;

			page->copyset = (1 << node);
			page->owner = node;
			DEBUG_DSM("[%d] owner for %lx is now %d\n", getpid(), pa, node);
		}
		else
		{
			/* serve a read copy */
			page->copyset |= (1 << node);
		}
	}

	ipc_send_fetch_reply(node, pa, is_write);
	mm_chmod(pa, is_write ? PROT_NONE : (PROT_READ|PROT_EXEC));
}

void
mm_handle_fetch_nak(int node, unsigned long pa, int is_write)
{
	struct request request;

	printf("[%d] FETCH NAK\n", cpu.id);
	usleep(10); /* BACKOFF */

	/* resend fetch request */
	request.type = IPC_FETCH;
	request.fetch.pa = pa;
	request.fetch.is_write = is_write;
	ipc_send(node, &request);
}

void
mm_handle_fetch_reply(unsigned long pa, int is_write)
{
	int manager;

	manager = pa >> NODE_SHIFT;
	if (manager == cpu.id)
	{
		struct managed_page *page = &managed_pages[(pa & NODE_MASK) >> PAGE_SHIFT];
		if (is_write)
			page->copyset = (1 << cpu.id);
		else
			page->copyset |= (1 << cpu.id);

		page->owner = cpu.id;
		DEBUG_DSM("[%d] owner for %lx is now %d\n", getpid(), pa, cpu.id);
	}

	mm_chmod(pa, is_write ? (PROT_READ|PROT_EXEC|PROT_WRITE) : (PROT_READ|PROT_EXEC));
}

void
mm_handle_invalidate_reply(unsigned long pa, int is_write)
{
	mm_handle_fetch_reply(pa, is_write);
}

void
mm_fetch(struct sigcontext *sc, unsigned long pa, int is_write, unsigned long address)
{
	struct request request;
	unsigned long copyset = 0;
	struct managed_page *page = NULL;
	int owner, manager;

	DEBUG_DSM("[%d] mm_fetch: ip=%lx va=%lx %s\n", getpid(), sc->sc_ip, address, is_write ? "write" : "read");

#if 0
	if (pa == last_invalidate_address)
	{
		printf("[%d] BACKING OFF\n", cpu.id);
		usleep(10);
		last_invalidate_address = 0;
	}
#endif

	owner = manager = pa >> NODE_SHIFT;
	if (manager == cpu.id)
	{
		page = &managed_pages[(pa & NODE_MASK) >> PAGE_SHIFT];
		owner = page->owner;
		copyset = page->copyset;
	}

	if (owner == cpu.id)
	{
		assert(manager == cpu.id);
		assert(is_write);

		DEBUG_DSM("[%d] invalidating all other copies\n", getpid());

		assert(page->invalidate_in_progress == 0);
		page->invalidate_in_progress = 1;
		request.type = IPC_INVALIDATE;
		request.invalidate.pa = pa;
		request.invalidate.is_write = is_write;
		ipc_broadcast(sc, copyset & ~(1 << cpu.id), &request);
		assert(page->invalidate_in_progress == 1);
		page->invalidate_in_progress = 0;

		DEBUG_DSM("[%d] invalidated\n", getpid());
	}
	else
	{
		request.type = IPC_FETCH;
		request.fetch.pa = pa;
		request.fetch.is_write = is_write;
		ipc_call(sc, owner, &request);
	}
}
#endif

void
mm_fault(struct sigcontext *sc, unsigned long address, unsigned long isr, int access_fault)
{
	unsigned long page = address & ~PAGE_MASK;
	struct pte *pte;
#ifdef DSM
	int is_exec = (isr & (1UL << 32)) ? 1 : 0;
#endif
	int is_write = (isr & (1UL << 33)) ? 1 : 0;

	DEBUG_MM("[%lx] mm_fault @ 0x%lx %d %d; ar.k6=%lx, ar.k7=%lx\n", sc->sc_ip, address, is_write, access_fault, cpu.cr[106], cpu.cr[107]);

	if (address >= 0xa000000000000000UL){
		fprintf(stderr, "Bad address %lx\n", address);
		vm_panic(sc, "bailing out\n");
	}
	if (address < pagesize)
	{
		/* NaT Page Consumption */
		cpu.cr[CR_IFA] = address;
		cpu.cr[CR_ITIR] = cpu.rr[address >> 61];
		deliver_exception(sc, NAT_CONSUMPTION, isr);
		return;
	}

#ifdef DSM
	if ((address > pagesize) && (address < MEM_SIZE))
	{
		mm_fetch(sc, page, is_write, address);
		return;
	}
#endif

	/*
	 * Special-case percpu area.
	 *
	 * Pretend the VHPT is turned off.
	 */
	if (!access_fault)
	{
	deliver:
		/* TLB miss */
		if (cpu.psr & PSR_IC)
		{
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[address >> 61];

			cpu.cr[CR_IHA] = thash(address);
			if (likely(address > CPUADDR && (cpu.rr[RGN(address)] & 1))) /* VHPT walker enabled */
				deliver_exception(sc, (isr & (1UL << 32)) ? ITLB_MISS : DTLB_MISS, isr);
			else
				deliver_exception(sc, (isr & (1UL << 32)) ? ALT_ITLB_MISS : ALT_DTLB_MISS, isr);
		}
		else
		{
			DEBUG_MM("data nested TLB miss.  R16(fault addr)=%lx, r29(saved b0)=%lx, r30(continuation)=%lx, r31(saved pr)=%lx\n",
				read_register(sc, 16),
				read_register(sc, 29),
				read_register(sc, 30),
				read_register(sc, 31));
			deliver_exception(sc, NESTED_DTLB_MISS, isr); /* Data Nested TLB vector */
		}
		return;
	}

	pte = pagetable[RGN(page)][HASH(page)];
	for (; pte != NULL; pte = pte->hash_next)
	{
		if (pte->va == page)
			goto found;
	}

	goto deliver;

found:
	DEBUG_MM("Found PTE va %lx->pa %lx", pte->va, pte->pa);
	if (!(pte->prot & PROT_ACCESSED))
	{
		DEBUG_MM(", Raising PROT_ACCESS fault\n");
		/* Access bit vector */
		cpu.cr[CR_IFA] = address;
		cpu.cr[CR_ITIR] = cpu.rr[RGN(page)];
		deliver_exception(sc, (isr & (1UL << 32)) ? 
				  I_ACCESS_BIT : D_ACCESS_BIT, isr);
		return;
	}

	if (is_write)
	{
		if (!(pte->prot & PROT_DIRTY))
		{
			DEBUG_MM("PROT_DIRTY");
			/* Dirty Bit vector */
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[RGN(page)];
			deliver_exception(sc, DIRTY_BIT_VECTOR, isr);
			return;
		}

		if (!(pte->prot & PROT_WRITE))
		{
			DEBUG_MM("ACCESS_RIGHTS");
			/* Instruction Access Rights/Data Access Rights vector */
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[RGN(page)];
			deliver_exception(sc, (isr & (1UL << 32)) ? 
					  I_ACCESS_RIGHTS : D_ACCESS_RIGHTS,
					  isr);
			return;
		
		}

	}
#ifdef DSM
	if (is_exec)
	{
		if (!(pte->prot & PROT_EXEC))
		{
			/* Instruction Access Rights vector */
			cpu.cr[CR_IFA] = address;
			cpu.cr[CR_ITIR] = cpu.rr[RGN(page)];
			deliver_exception(sc, I_ACCESS_RIGHTS, isr);
			return;
		}
	}

	mm_fetch(sc, pte->pa & ~PAGE_MASK, is_write, address);
#endif
}


void *
mmapall(void *start, size_t length, int prot, int flags,
                  int fd, off_t offset)
{
#ifdef OPTIMISE_USE_MULTI_AS
	int i;
	if (ncontexts > 1)
		for (i = (last_asid + 1) % ncontexts; i != last_asid; i = (i + 1)%ncontexts) {
		switch_as(i);
		if (mmap(start, length, prot, flags, fd, offset) == MAP_FAILED) {
			switch_as(last_asid);
			return MAP_FAILED;
		}
	}
	switch_as(last_asid);
#endif
	return mmap(start, length, prot, flags, fd, offset);
}

void
as_init(void)
{
	int i;
	int j = -1;
	while ((i = new_as()) >= 0 && i < NCONTEXTS)
		j = i;
	ncontexts = j + 1;
	fprintf(stderr, "Kernel allows %d address space contexts\n", ncontexts);
#ifndef OPTIMISE_USE_MULTI_AS
	ncontexts = NCONTEXTS;
#endif
	asid_hist = ncontexts * 9 / 10;
	fprintf(stderr, "asid_hist = %d\n", asid_hist);
	if (j == -1) {
		ncontexts = NCONTEXTS;
		asid_hist = ASID_HIST;
	}
}

#if DEBUG & DEBUG_CTXT_BIT
void
dump_ctx(void)
{
	int i, j;

	for (j = 0; j < HI_REGION; j++) {
		fprintf(stderr, "Region %d: ", j);
		for (i = 0; i < ncontexts; i++)
			fprintf(stderr, " %d ", asid[j][i]);
		fprintf(stderr, "\n");
	}
	fprintf(stderr, "\n Hist:\n");
	for (j = 0; j < HI_REGION; j++) {
		fprintf(stderr, "Region %d: ", j);
		for (i = 0; i < asid_hist; i++)
			fprintf(stderr, " %d ", context_hist[j][i]);
		fprintf(stderr, "\n");
	}
}
#endif
