#define _GNU_SOURCE
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <asm/break.h>
#include <sys/ptrace.h>
#include <sched.h>
#include "vmm-offsets.h"
#include "vm.h"
#include "proto.h"
#include "debug.h"

#define PTRACE_FPH_TRAP 0x26
extern int in_panic;
unsigned long cpu_freq = CPU_FREQ;


unsigned long *
find_register(struct sigcontext *sc, unsigned int regno);

extern unsigned long pagesize;
extern int keyboard_irq, network_irq, network_fd, simscsi_irq;

unsigned long thread_pointer;

#define save_tp(x)	asm("mov %0 = r13" : "=r"(x));
#define reload_tp(x)	asm("mov r13 = %0" : : "r"(x));

static void
handle_SIGILL(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;

	INC_STAT(sigill);

	reload_tp(thread_pointer);
	DEBUG_EXCEPTION("SIGILL @%lx\n", sc->sc_ip);

	emu_priv(sc);
}

static void
handle_SIGSEGV(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	unsigned long address = (unsigned long)(info->si_addr);
	extern void _text, _etext;
	reload_tp(thread_pointer);

	INC_STAT(faults);

	DEBUG_EXCEPTION("SIGSEGV: 0x%lx VA 0x%lx code=%d\n", sc->sc_ip, address, info->si_code);

	if ((sc->sc_ip >= (unsigned long)&_text) && (sc->sc_ip < (unsigned long)&_etext))
	{
		if (dumping)
			longjmp(continuation, 1);
		fprintf(stderr, "[%d] SIGSEGV in VMM: IP 0x%lx VA 0x%lx\n", getpid(), sc->sc_ip, address);
		abort();
		vm_panic(sc, "SIGSEGV");
	}
	if (sc->sc_ip >= (5UL<<61)) {
		/*
		 * Catch the sigtramp write into the rnat slot, where
		 * the RBS has been invalidated
		 */
		sc->sc_ip++;
		return;
	}
	
		
#ifdef SMP
	if ((address >= PIB_BASE) && (address < PIB_BASE + pagesize))
	{
		emu_store_PIB(sc, address);
		return;
	}
#endif

	mm_fault(sc, address, info->si_isr, (info->si_code == SEGV_ACCERR));
}

static void
handle_SIGALRM(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	reload_tp(thread_pointer);
	DEBUG_EXCEPTION("SIGALRM @ %p\n", (void *)sc->sc_ip);
	INC_STAT(ticks);

#ifdef DSM
	process_request_queue(sc);
#endif
	fflush(stdout);
	deliver_interrupt(sc, cpu.cr[CR_ITV] & 0xff);
}

int
handle_SIGIO_core(struct siginfo *info, struct sigcontext *sc)
{
	DEBUG_EXCEPTION("SIGIO signal=%d, fd=%d code=%d, band=%ld\n", info->si_signo, info->si_fd, info->si_code, info->si_band);
	INC_STAT(sigio);

	switch (info->si_signo) {
	case SIGSCSI:
		if (simscsi_irq) {
			deliver_interrupt(sc, simscsi_irq); 
			break;
		}
		/* FALL_THROUGH */
	default:	
		DEBUG_EXCEPTION("SIGIO: Unexpected signal\n");
		return 0;

	case SIGKEYBD:
		if (info->si_code == POLL_IN)
			deliver_interrupt(sc, keyboard_irq);
		break;

	case SIGNETWK:
		deliver_interrupt(sc, network_irq);
		break;

	case SIGIPC:
#ifdef SMP
		return ipc_handle_async(sc, info->si_fd);
#else
		printf("Unexpected SIGIPC\n");
#endif
		break;
	}

	return 0;
}

static void
handle_SIGIO(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	reload_tp(thread_pointer);

	handle_SIGIO_core(info, sc);
}

/*
 * One gets here after the ptrace thread has redirected a system call.
 */
static void
handle_SIGUSR2(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	reload_tp(thread_pointer);
	INC_STAT(redirects);
	DEBUG_EXCEPTION("SIGUSR2 @ %lx, imm==%x\n", sc->sc_ip, info->si_imm);
	DEBUG_EXCEPTION("R15=%lu\n", sc->sc_gr[15]);
	cpu.cr[CR_IIM] = __IA64_BREAK_SYSCALL;
	deliver_exception(sc, BREAK_EXCEPTION, 0);
}

const char*
decode_trap(int code)
{
	const char * trapnames[] = {
		"SSC",
		"COVER",
		"HALT",
		"0x80004",
		"0x80005",
		"TLBOP",
		"WRITE_RR",
		"SET_DFH",
		"SSM_PSR_I",
		"0x8000A",
		"0x8000B",
		"0x8000C",
		"0x8000D",
		"0x8000E",
		"0x8000F",
		"BSW_0",
		"BSW_1",
		"READ_IVR",
		"WRITE_ITM",
		"THASH",
		"TPA",
		"WRITE_PSR_L",
		"0x80017",
		"0x80018",
		"0x80019",
		"0x8001A",
		"0x8001B",
		"0x8001C",
		"0x8001D",
		"0x8001E",
		"0x8001F",
		"RFI",
		"DELIVER_IRQ",
		"PAUSE",
	};
	if (code >= BREAK_SSC && code-BREAK_SSC < sizeof trapnames/sizeof trapnames[0])
		return trapnames[code - BREAK_SSC];
	if (code == __IA64_BREAK_SYSCALL)
		return "syscall";
	return "unrecognised";
}

/*
 * One gets here after a TRAP (i.e., a break.i or similar instruction)
 */
static void
handle_SIGTRAP(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	static int can_ptrace_fph = 1;

	reload_tp(thread_pointer);

	unsigned long ip = sc->sc_ip;
	unsigned int slot = ip & 0xf;
	long arg0, arg1, arg2, arg3;
	siginfo_t newinfo;
	sigset_t sigset;

	sc->sc_ip += (slot == 2) ? 0xe : 1;

	INC_STAT(breaks);

	if (info->si_imm != BREAK_SSC) {

		DEBUG_EXCEPTION("SIGTRAP, @%lx imm = %s (r4=%lx)\n", sc->sc_ip, decode_trap(info->si_imm), read_register(sc, 4));
	}

	switch (info->si_imm)
	{
	case 0: /* panic */
		vm_panic(sc, "panic");
		break;

	case BREAK_SSC: /* ssc */
		INC_STAT(ssc);

		arg0 = (long)read_register(sc, 32);
		arg1 = (long)read_register(sc, 33);
		arg2 = (long)read_register(sc, 34);
		arg3 = (long)read_register(sc, 35);
		sc->sc_gr[8] = do_ssc(sc, arg0, arg1, arg2, arg3, sc->sc_gr[15]);
		break;

	case BREAK_COVER: /* cover */
		INC_STAT(break_cover);
		emu_cover(sc);
		break;

	case BREAK_SET_DFH:
		INC_STAT(break_dfh);

		cpu.psr |= PSR_DFH;
		if (can_ptrace_fph && (ptrace(PTRACE_FPH_TRAP) == -1)) {
			vm_perror("ptrace_FPH_TRAP\n");
			can_ptrace_fph = 0;
		}
		break;

	case BREAK_HALT: /* halt */
		INC_STAT(break_halt);
#if 0
		fprintf(stderr, "[%d] halt\n", getpid());
#endif
		/*
		 * We could have missed an interrupt here.
		 * Check!
		 */
		if (cpu.irq_pending) {
			check_interrupts(sc);
			break;
		}


		sigemptyset(&sigset);
		sigaddset(&sigset, SIGALRM);
		sigaddset(&sigset, SIGIPC);
		sigaddset(&sigset, SIGKEYBD);
		sigaddset(&sigset, SIGSCSI);
		sigaddset(&sigset, SIGNETWK);
		sigaddset(&sigset, SIGUSR2);
		sig = sigwaitinfo(&sigset, &newinfo);
		switch (sig)
		{
		case -1:
			vm_perror("sigwaitinfo");
			break;

		case SIGALRM:
			handle_SIGALRM(sig, &newinfo, uc);
			break;

		case SIGKEYBD:
		case SIGNETWK:
		case SIGIPC:
		case SIGSCSI:
		case SIGIO:
			handle_SIGIO(sig, &newinfo, uc);
			break;
		}
#if 0
		fprintf(stderr, "[%d] wakeup\n", getpid());
#endif
		break;

	case __IA64_BREAK_SYSCALL:
		INC_STAT(break_syscall);

		cpu.cr[CR_IIM] = info->si_imm;
		deliver_exception(sc, BREAK_EXCEPTION, 0);
		break;
		

	case BREAK_TPA:
		INC_STAT(break_tpa);
		emu_tpa(sc, 5, 5);
		break;

	case BREAK_TLBOP:
		INC_STAT(break_tlbop);
		tlbop(sc, read_register(sc, 5), 6, 7);
		break;

	case BREAK_WRITE_RR:
		INC_STAT(break_rr);
		emu_write_rr(sc, 5, 6);
		break;

	case BREAK_WRITE_ITM:
		INC_STAT(break_itm);

		cpu.cr[CR_ITM] = read_register(sc, 5);
		set_timer();
		break;

	case BREAK_RFI:
		INC_STAT(break_rfi);
 		sc->sc_ip = (cpu.cr[CR_IIP] & ~0xf) | (BITS(cpu.cr[CR_IPSR], PSR_RI_OFFSET, 2));
		if (cpu.cr[CR_IFS] & CFM_VALID)
			sc->sc_cfm = BITS(cpu.cr[CR_IFS], 0, 38);
		cpu.cfle = 1;
		write_psr(sc, cpu.cr[CR_IPSR] & ~PSR_RI);
		break;

	case BREAK_DELIVER_IRQ:
		INC_STAT(break_irq);

		check_interrupts(sc);
		break;

	case BREAK_READ_IVR:
		INC_STAT(break_ivr);

		DEBUG_EXCEPTION("READ_IVR: %lx\n", cpu.cr[CR_IVR]);
		write_register(sc, 5, cpu.cr[CR_IVR]);
		cpu.in_service_irq = cpu.cr[CR_IVR];
		clear_interrupt(cpu.cr[CR_IVR]);
		break;

	case BREAK_DEBUG:
		INC_STAT(break_debug);

		fprintf(stderr, "\n\nBREAK_DEBUG\n");
		dump_cpu();
		dump_regs(sc);
		dump_maps();
		fprintf(stderr, "\n************\n\n");
		break;

	case BREAK_PAUSE:
		INC_STAT(break_pause);
#if 0
		{
			struct timespec ts;
			ts.tv_sec = 0;
			ts.tv_nsec = 1000; /* sleep >= 1usec; effectively 
					      give up rest of timeslice */
			nanosleep(&ts, NULL);
		}
#else
		sched_yield();
#endif
		break;

	default:
		/*
		* I want to know about these ... panic the virtual machine.
		*/
		fprintf(stderr, "TRAP %x\n", info->si_imm);
		vm_panic(sc, "TRAP");
#if 0
		/*
		 * `Proper' behaviour is to send the trap on to the guest.
		 */
		cpu.cr[CR_IIM] = info->si_imm;
		deliver_exception(sc, BREAK_EXCEPTION, 0);
#endif
		break;
	}
}

static void
handle_SIGFPE(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	reload_tp(thread_pointer);
	DEBUG_EXCEPTION("SIGFPE\n");
	INC_STAT(fpe);

//	if (cpu.psr & PSR_DFH)
	{
		/* Disabled FP register vector */
		deliver_exception(sc, DISABLED_FPH, 2);
	}
}

static void
handle_SIGTERM(int sig, siginfo_t *info, void *uc)
{
	reload_tp(thread_pointer);
	DEBUG_EXCEPTION("SIGTERM\n");

	fprintf(stderr, "Received SIGTERM, shutting down...\n");
	vm_shutdown();
}

static void
handle_SIGHUP(int sig, siginfo_t *info, void *uc)
{
	struct sigcontext *sc = &((struct ucontext *)uc)->uc_mcontext;
	reload_tp(thread_pointer);
	dump_regs(sc);
	dump_maps();
	dump_stats();
}

void
set_timer()
{
	unsigned long target_time = cpu.cr[CR_ITM];
	unsigned long current_time, microsec;
	struct itimerval val;

	static unsigned long lastmsg;

	asm("mov %0=ar.itc" : "=r"(current_time));

	if (target_time < current_time)
	{
		if (current_time > lastmsg + cpu_freq) {
			printf("set_timer: target time %lX before current time %lX using %lX\n",
		       target_time, current_time, current_time + cpu_freq/1000);
			lastmsg = current_time;
		}
		/* 1 millisec further on */
		target_time = current_time + 1*cpu_freq/1000;
	}

	microsec = (target_time - current_time + (cpu_freq/2000000)) / (cpu_freq/1000000);
	val.it_value.tv_sec = microsec / 1000000;
	val.it_value.tv_usec = microsec % 1000000;
	val.it_interval.tv_sec = val.it_interval.tv_usec = 0;

	if (setitimer(ITIMER_REAL, &val, NULL) == -1)
		vm_perror("setitimer");
}

void
signal_init(void)
{
	static double stack[SIGSTKSZ/sizeof(double)]; /* ensure 16 byte alignment */
	struct sigaction sa;
	stack_t newstack;

	newstack.ss_sp = stack;
	newstack.ss_flags = 0;
	newstack.ss_size = sizeof(stack);
	if (sigaltstack(&newstack, NULL) == -1)
		return;

	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGALRM);
	sigaddset(&sa.sa_mask, SIGIO);
	sigaddset(&sa.sa_mask, SIGIPC);
	sigaddset(&sa.sa_mask, SIGKEYBD);
	sigaddset(&sa.sa_mask, SIGSCSI);
	sigaddset(&sa.sa_mask, SIGNETWK);
	
	sigaddset(&sa.sa_mask, SIGUSR2);
       	sa.sa_flags = SA_SIGINFO | SA_ONSTACK;

	sa.sa_sigaction = handle_SIGILL;
	if (sigaction(SIGILL, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGSEGV;
	if (sigaction(SIGSEGV, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGALRM;
	if (sigaction(SIGALRM, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGIO;
	if (sigaction(SIGIO, &sa, NULL) == -1)
		vm_perror("sigaction");
	if (sigaction(SIGIPC, &sa, NULL) == -1)
		vm_perror("sigaction");
	if (sigaction(SIGKEYBD, &sa, NULL) == -1)
		vm_perror("sigaction");
	if (sigaction(SIGNETWK, &sa, NULL) == -1)
		vm_perror("sigaction");
	if (sigaction(SIGSCSI, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGTRAP;
	if (sigaction(SIGTRAP, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGUSR2;
	if (sigaction(SIGUSR2, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGHUP;
	if (sigaction(SIGHUP, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGTERM;
	if (sigaction(SIGTERM, &sa, NULL) == -1)
		vm_perror("sigaction");

	sa.sa_sigaction = handle_SIGFPE;
	if (sigaction(SIGFPE, &sa, NULL) == -1)
		vm_perror("sigaction");

	save_tp(thread_pointer);
}
