#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <errno.h>
#define __ASSEMBLY__
#include <asm/ptrace.h>
#include <asm/ptrace_offsets.h>
#undef __ASSEMBLY__
#include <asm/unistd.h>
#include "vm.h"
#include "debug.h"

#ifndef PTRACE_SETOPTIONS
#define PTRACE_SETOPTIONS	0x4200
#define PTRACE_O_TRACESYSGOOD	0x00000001
#endif
#define PTRACE_O_TRACESYSEXIT    0x00000080
#ifdef OPTIMISE_PTRACE_SIGNALS
#define PTRACE_O_NOSIGSTOP	0x00000100
#else
#define PTRACE_O_NOSIGSTOP 0
#endif
#ifndef PTRACE_ADDRRANGE
#define PTRACE_ADDRRANGE 22
#endif
#ifndef PTRACE_VIRTUAL
#  define PTRACE_VIRTUAL 39
#endif

#if DEBUG & DEBUG_STRACE_BIT
#define DEBUG_STRACE(args...) fprintf(stderr, args);
#else
#define DEBUG_STRACE(...) /* nothing */
#endif

unsigned long redirects;
unsigned long VMM;
unsigned long sigreturns;
unsigned long totalstops;
unsigned long signalstop;
unsigned long sigstop;
unsigned long sigtrap;

static void
dump_pt_regs(struct pt_all_user_regs *regs)
{
	int i;

	printf("ip   = %016lx\tpsr  = %016lx\tcfle = %d\n", regs->cr_iip, regs->cr_ipsr, cpu.cfle);
	printf("cfm  = %016lx\tpfs  = %016lx\n", regs->cfm, regs->ar[PT_AUR_PFS]);
	printf("bsp  = %016lx\trsc  = %016lx\n", regs->ar[PT_AUR_BSP], regs->ar[PT_AUR_RSC]);
	printf("pr   = %016lx\tnat  = %016lx\n", regs->pr, regs->nat);
	printf("rnat = %016lx\tunat = %016lx\n", regs->ar[PT_AUR_RNAT], regs->ar[PT_AUR_UNAT]);
	for (i = 0; i < 32; i += 2)
		printf("r%-2d  = %016lx\tr%-2d  = %016lx\n", i, regs->gr[i], i+1, regs->gr[i+1]);
	for (i = 0; i < 8; i += 2)
		printf("b%-2d  = %016lx\tb%-2d  = %016lx\n", i, regs->br[i], i+1, regs->br[i+1]);
	for (i = 0; i < 32; i += 2)
		fprintf(stderr, "f%-2d  = %016lx%016lx\tf%-2d  = %016lx%016lx\n", i, regs->fr[i].u.bits[1], regs->fr[i].u.bits[0], i+1, regs->fr[i+1].u.bits[1], regs->fr[i+1].u.bits[0]);
	printf("dcr  = %016lx\titm  = %016lx\n", cpu.cr[CR_DCR], cpu.cr[CR_ITM]);
	printf("iva  = %016lx\tpta  = %016lx\n", cpu.cr[CR_IVA], cpu.cr[CR_PTA]);
	printf("ipsr = %016lx\tisr  = %016lx\n", cpu.cr[CR_IPSR], cpu.cr[CR_ISR]);
	printf("iip  = %016lx\tifa  = %016lx\n", cpu.cr[CR_IIP], cpu.cr[CR_IFA]);
	printf("itir = %016lx\tiipa = %016lx\n", cpu.cr[CR_ITIR], cpu.cr[CR_IIPA]);
	printf("ifs  = %016lx\tiim  = %016lx\n", cpu.cr[CR_IFS], cpu.cr[CR_IIM]);
	printf("iha  = %016lx\n", cpu.cr[CR_IHA]);
}

static int
ptrace_debug(pid_t pid)
{
	char line[256];
	unsigned long address, value;

	while (1)
	{
		printf("> ");
		fflush(stdout);
		if (fgets(line, sizeof(line), stdin) == NULL)
			return 0;
		if (!line[0])
			continue;
		switch (line[0])
		{
			case 'c':
				printf("continue\n");
				return 0;

			case 's':
				printf("singlestep\n");
				return 1;

			case 'd':
				if (line[1] != ' ')
					continue;

				address = strtoul(&line[2], NULL, 16);
				value = ptrace(PTRACE_PEEKDATA, pid, address, NULL);
				printf("%lx = %lx\n", address, value);
				break;
		}
	}
}

/*
 * The VMM runs in the same process as the guest.
 * Thus we need to redirect system calls that 
 * are for the guest to the guest, but allow the VMMs system 
 * calls to reach the host OS.
 *
 * This routine intercepts all system calls.  If the IP is 
 * between _start and _end, or in the gate page, the code is continued.
 * Otherwise, the syscall number is replaced with 
 * something harmless (getpid()), and continued, then on exit from the
 * syscall, a SIGUSR2 is delivered to the traced thread.
 * The VMM converts this into an exception to deliver to the guest OS.
 */
#define WSYSCALLENTRY(status) (!((status) & 0x10000U))
void
syscall_trace(void)
{
	pid_t pid, pids[64];
	long in_syscall[64];
	int status, signal_no;
	unsigned long pc;
	int i;
	siginfo_t siginfo;
	int singlestep;
	extern void _text, _etext, _end;


#ifdef OPTIMISE_DISABLE_PTRACE
	fprintf(stderr, "Child %u: calling PTRACE_VIRTUAL\n", getpid());
	if (ptrace(PTRACE_VIRTUAL, getpid(), &_text, &_end-&_text) == 0) {
		return;
	}
#endif
	/*
	 * If PTRACE_VIRTUAL works,
	 * don't do this
	 */
	switch (pid = vm_clone())
	{
		case -1:
			vm_perror("fork");
			return;

		case 0: /* child */
			ptrace(PTRACE_TRACEME);
			kill(getpid(), SIGSTOP);
			return;
	}

	fprintf(stderr, "[%d] Hi, I am the trace thread!\n", getpid());
	for (i = 0; i < 64; i++) {
		in_syscall[i] = 0;
		pids[i] = 0;
	}
#ifdef OPTIMISE_SET_ADDRRANGE

	if (ptrace(PTRACE_ADDRRANGE, pid, 0UL, &_text - 1) == -1)
		fprintf(stderr, "ptrace(ADDRRANGE failed: %s\n", sys_errlist[errno]);
	else
		ptrace(PTRACE_ADDRRANGE, pid, &_etext, (5UL<<61)-(unsigned long)&_etext);
#endif

	while ((pid = waitpid(-1, &status, WUNTRACED)) != -1)
	{
		if (!WIFSTOPPED(status))
		{
			if (WIFSIGNALED(status))
				printf("Exited with signal %d\n", WTERMSIG(status));
			else
				printf("Exited with status %d\n", WEXITSTATUS(status));
			fprintf(stderr, "%lu stops, %lu redirects, %lu VMM, %lu sigreturns, %lu signalstops, %lu traps, %lu sigstops\n",
				totalstops, redirects, VMM, sigreturns, signalstop, sigtrap, sigstop);
			break;
		}

		totalstops++;

		signal_no = WSTOPSIG(status);
		singlestep = 0;

		if (signal_no == (SIGTRAP | 0x80))
		{
			// GRRR - kernel should tell us whether syscall entry or exit
			for (i = 0; i < 64; i++)
			{
				if (pids[i] == pid)
					break;
				else if (pids[i] == 0)
				{
					pids[i] = pid;
					break;
				}
			}

			if (in_syscall[i] == 0)
			{
				unsigned long syscall;

				pc = ptrace(PTRACE_PEEKUSER, pid, PT_CR_IIP, NULL);
				syscall = ptrace(PTRACE_PEEKUSER, pid, PT_R15, NULL);


				/* 
				 * common case is sigreturn from gate page (rgn 5) 
				 * Should never see this if addrange is working.
				 */
				if (((pc >> 61) == 5) || ((pc >= (unsigned long)&_text) && (pc < (unsigned long)&_etext)))
				{
					
					in_syscall[i] = -1;
					if ((pc >> 61) == 5)
						sigreturns++;
					else
						VMM++;
				}
				else if (WSYSCALLENTRY(status))
				{
					DEBUG_STRACE("[%d] [%lx] SYSCALL %ld\n", pid, pc, syscall);
					ptrace(PTRACE_POKEUSER, pid, PT_R15, __NR_getpid);
					in_syscall[i] = syscall;
					redirects++;
				}
				else
					sigreturns++;
				signal_no = 0;
			}
			else if (in_syscall[i] == -1)
			{
				/* return from VMM syscall */
				in_syscall[i] = 0;
				signal_no = 0;
			}
			else
			{
				/* return from redirected syscall - restore syscall number */
				ptrace(PTRACE_POKEUSER, pid, PT_R15, in_syscall[i]);
				in_syscall[i] = 0;
				signal_no = SIGUSR2;
			}
		}
		else if (signal_no == SIGSTOP)
		{
			sigstop++;
			if (ptrace(PTRACE_SETOPTIONS, pid, NULL, PTRACE_O_TRACESYSEXIT|PTRACE_O_TRACESYSGOOD|PTRACE_O_NOSIGSTOP) == -1) {
				if (ptrace(PTRACE_SETOPTIONS, pid, NULL, PTRACE_O_TRACESYSGOOD) == -1) {				
					vm_perror("PTRACE_SETOPTIONS");
				}
			}

#if 0
			ptrace(PTRACE_POKEUSER, pid, PT_CR_IPSR, ptrace(PTRACE_PEEKUSER, pid, PT_CR_IPSR) | PSR_DB);
#endif

			/* ptrace_debug(pid); */
#if 0
			if (ptrace(PTRACE_POKEUSER, pid, PT_DBR, /* 0x7f24570 */ 0x2000000000042570) == -1)
				vm_perror("PT_DBR");
			if (ptrace(PTRACE_POKEUSER, pid, PT_DBR+8, 0xfffffffffffffff0) == -1)
				vm_perror("PT_DBR+8");
#endif
#if 0
			if (ptrace(PTRACE_POKEUSER, pid, PT_IBR, 0x2000000000025350) == -1)
				vm_perror("PT_IBR");
			if (ptrace(PTRACE_POKEUSER, pid, PT_IBR+8, 0xffffffffffffffff) == -1)
				vm_perror("PT_IBR+8");
#endif
			signal_no = 0;
		}
#if 0
		else if (signal_no == SIGILL)
		{
			extern void _text, _etext;
			struct pt_all_user_regs regs;
			long pc, code;
			sigill++;

			if (ptrace(PTRACE_GETSIGINFO, pid, NULL, &siginfo) == -1)
				vm_perror("PTRACE_GETSIGINFO");
			if (ptrace(PTRACE_GETREGS, pid, NULL, &regs) == -1)
				vm_perror("PTRACE_GETREGS");
			pc = regs.cr_iip;
			code = 1;
#if 1
			if ((((pc >= (unsigned long)&_text) && (pc < (unsigned long)&_etext)))
				&& (code = ptrace(PTRACE_PEEKTEXT, pid, (void *)pc, NULL)) == -1)
				vm_perror("PTRACE_PEEKTEXT");
#endif
			if (verbose)
				fprintf(stderr, "SIGILL ip = %lx addr = %lx rsc = %lx bsp = %lx bspstore = %lx code = %x vector = %x isr = %lx code = %lx\n", regs.cr_iip, siginfo.si_addr, regs.ar[PT_AUR_RSC], regs.ar[PT_AUR_BSP], regs.ar[PT_AUR_BSPSTORE], siginfo.si_code, siginfo.si_imm, siginfo.si_isr, code);
		}
		else if (signal_no == SIGSEGV)
		{
			struct pt_all_user_regs regs;
			sigsegv++;
			pc = ptrace(PTRACE_PEEKUSER, pid, PT_CR_IIP, NULL);
			if (pc == 0x200000000507A630)
			{
			if (ptrace(PTRACE_GETSIGINFO, pid, NULL, &siginfo) == -1)
				vm_perror("PTRACE_GETSIGINFO");
			if (ptrace(PTRACE_GETREGS, pid, NULL, &regs) == -1)
				vm_perror("PTRACE_GETREGS");
			fprintf(stderr, "SIGSEGV\n");
			dump_pt_regs(&regs);
			}
		}
#endif /* 0 */
		else if (signal_no == SIGTRAP)
		{
			DEBUG_STRACE("strace:  got SIGTRAP\n");
			sigtrap++;
			if (ptrace(PTRACE_GETSIGINFO, pid, NULL, &siginfo) == -1)
				vm_perror("PTRACE_GETSIGINFO");
			if (siginfo.si_code != TRAP_BRKPT)
			{
				struct pt_all_user_regs regs;
				if (ptrace(PTRACE_GETREGS, pid, NULL, &regs) == -1)
					 vm_perror("PTRACE_GETREGS");
				pc = regs.cr_iip /* ptrace(PTRACE_PEEKUSER, pid, PT_CR_IIP, NULL) */;
				printf("HARDWARE BREAKPOINT @ %lx, si_code=%d\n", pc, siginfo.si_code);
				dump_pt_regs(&regs);
				singlestep = ptrace_debug(pid);
				ptrace(PTRACE_POKEUSER, pid, PT_CR_IPSR, ptrace(PTRACE_PEEKUSER, pid, PT_CR_IPSR) | PSR_ID | PSR_DD);
				signal_no = 0;
			}
		}
		else
		{
			signalstop++;
#if DEBUG & DEBUG_STRACE_BIT
			pc = ptrace(PTRACE_PEEKUSER, pid, PT_CR_IIP, NULL);
			DEBUG_STRACE("[%d] [%lx] SIGNAL %d: %s\n", pid, pc, signal_no, strsignal(signal_no));
#endif
		}

		ptrace(singlestep ? PTRACE_SINGLESTEP : PTRACE_SYSCALL, pid, NULL, signal_no);
	}

	exit(0);
}

